Package: MCseqReplic
Version: 1.0.0
Date: 2026-04-29
Title: Monte Carlo Simulations of Time Changes in Sequences
Authors@R: c(
  person("Gilbert", "Ritschard",
    role = c("aut", "cre", "cph"),
    email = "gilbert.ritschard@unige.ch",
    comment = c(ORCID = "0000-0001-7776-0903")))
Depends: R (>= 3.0.0), TraMineR (>= 2.2-2)
Imports: WeightedCluster (>= 1.6.0), aricode, doParallel, foreach,
        doSNOW, iterators, stats, vegan, wCorr
Description: Generates replicated sets of sequences with Monte Carlo simulated timing changes and computes various indicators for evaluating effects of timing uncertainty on sequence analysis results. See Ritschard, G. and Liao, T.F. (2026): "Assessing the Impact of Timing Errors in Sequence Analysis". International Journal of Social Research Methodology <doi:10.1080/13645579.2026.2666297>.
License: GPL (>= 2)
URL: http://traminer.unige.ch
Encoding: UTF-8
Maintainer: Gilbert Ritschard <gilbert.ritschard@unige.ch>
RoxygenNote: 7.3.2
NeedsCompilation: no
Packaged: 2026-04-29 07:20:07 UTC; grits
Author: Gilbert Ritschard [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-7776-0903>)
Repository: CRAN
Date/Publication: 2026-04-29 19:00:13 UTC
Built: R 4.6.0; ; 2026-05-01 21:34:29 UTC; unix

Package: MCseqReplic
Version: 1.0.0
Date: 2026-04-29
Title: Monte Carlo Simulations of Time Changes in Sequences
Authors@R: c(
  person("Gilbert", "Ritschard",
    role = c("aut", "cre", "cph"),
    email = "gilbert.ritschard@unige.ch",
    comment = c(ORCID = "0000-0001-7776-0903")))
Depends: R (>= 3.0.0), TraMineR (>= 2.2-2)
Imports: WeightedCluster (>= 1.6.0), aricode, doParallel, foreach,
        doSNOW, iterators, stats, vegan, wCorr
Description: Generates replicated sets of sequences with Monte Carlo simulated timing changes and computes various indicators for evaluating effects of timing uncertainty on sequence analysis results. See Ritschard, G. and Liao, T.F. (2026): "Assessing the Impact of Timing Errors in Sequence Analysis". International Journal of Social Research Methodology <doi:10.1080/13645579.2026.2666297>.
License: GPL (>= 2)
URL: http://traminer.unige.ch
Encoding: UTF-8
Maintainer: Gilbert Ritschard <gilbert.ritschard@unige.ch>
RoxygenNote: 7.3.2
NeedsCompilation: no
Packaged: 2026-04-29 07:20:07 UTC; grits
Author: Gilbert Ritschard [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-7776-0903>)
Repository: CRAN
Date/Publication: 2026-04-29 19:00:13 UTC
Built: R 4.5.2; ; 2026-05-01 18:19:16 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,MCratios)
S3method(print,distMC)
S3method(summary,MCratios)
S3method(summary,distMC)
export(MCExtractDist)
export(MCclustcomp)
export(MCclustqual)
export(MCdisscorr)
export(MCdisslist)
export(MCmdscorr)
export(MCnunique)
export(MCpj)
export(MCratios)
export(MCseqReplicate)
export(MCseqdistSE)
export(MCudist)
export(seqdistMCSE)
import(TraMineR)
importFrom(TraMineR,seqdist)
importFrom(WeightedCluster,as.clustrange)
importFrom(WeightedCluster,wcAggregateCases)
importFrom(WeightedCluster,wcKMedRange)
importFrom(aricode,clustComp)
importFrom(doParallel,registerDoParallel)
importFrom(doSNOW,registerDoSNOW)
importFrom(foreach,"%:%")
importFrom(foreach,"%do%")
importFrom(foreach,"%dopar%")
importFrom(foreach,foreach)
importFrom(iterators,icount)
importFrom(parallel,detectCores)
importFrom(parallel,makeCluster)
importFrom(parallel,makePSOCKcluster)
importFrom(parallel,stopCluster)
importFrom(stats,as.dist)
importFrom(stats,dpois)
importFrom(stats,hclust)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,uniroot)
importFrom(utils,packageVersion)
importFrom(utils,setTxtProgressBar)
importFrom(utils,txtProgressBar)
importFrom(vegan,wcmdscale)
importFrom(wCorr,weightedCorr)

#' @title Generate distribution of timing errors
#'
#' @description
#' Generates a distribution of timing errors that complies with the provided expected size of non-zero timing errors and the expected probability of no error.
#'
#' @details
#' Currently \code{\link{MCseqReplicate}} expects a vector Pj with same number of backward and forward error values. To comply with this, the shorter side of Pj is by default filled with zeros.
#'
#'
#'
#' @param Emean scalar or vector of size two. Expected size of non-zero timing errors. If a vector, the first value is used for negative errors and the second value for positive errors. If a scalar, the value is used for both negative and positive errors. Values must be strictly greater than 1.
#' @param pzero number in range [0,1]. Probability of no-error. If \code{NULL} (default), \code{pzero} is set to the the greatest probability of zero between the right and left side Poisson distributions.
#' @param maxterr integer. Maximal error size to consider. Default is 10.
#' @param pinterv control value used for solving numerically an implicit function. Default is .99 and should be increased in case the zero of the implicit function cannot be found because of ending values of same sign.
#' @param fill.short.side logical. Should the shortest side be filled with zeros to equal length of the other side. Default is \code{TRUE}.
#'
#' @returns The vector of probabilities Pj with the computed \code{lambda} values as attribute.
#'
#' @seealso \code{\link{MCseqReplicate}}
#'
#' @export
#'
#' @importFrom stats dpois uniroot
#'
#' @examples
#' # expected timing error of 1.2 on each side
#' MCpj(Emean=1.2, pzero=.4)
#'
#' # expected backward timing error higher than for forward errors
#' MCpj(Emean=c(3.5,1.2), pzero=.4)
#'
#'

MCpj <- function(Emean, pzero=NULL, maxterr=10, pinterv=.99, fill.short.side=TRUE){

  if (length(Emean) > 2){
    warning("length of Emean greater than 2, only first 2 values are used!")
    vEmean <- Emean[1:2]
  }

  if (length(Emean)==2){
    nEmean <- 2
  } else {
    nEmean <- 1
  }

  lpois <- list()
  vEmean <- Emean
  pz <- 0
  lbda <- NULL
  for (i in 1:nEmean) {
    Emean <- vEmean[i]

    if (Emean <= 1){
      stop("Emean must be strictly greater than 1")
    }

    # Define the implicit function F(x, y) = 0
    implicit_fun <- function(lambda, expected) {
      return(expected - lambda/(1-exp(-lambda)))
    }

    # Find the root for lambda in the interval around the expected value
    interv <- c(Emean - Emean * pinterv, Emean + Emean * pinterv)
    # uniroot() requires the function to be defined such that the interval endpoints
    # produce values with opposite signs.

    tryCatch({
      solution_lambda <- uniroot(f = implicit_fun, interval = interv, expected = Emean)
    }, error = function(e) {
      print(paste("Error finding root:", e$message))
    })

    lambda <- max(solution_lambda$root,.001)
    pois <- dpois(1:maxterr, lambda)
    pz <- max(pz, dpois(0, lambda))
    lbda <- c(lbda, lambda)

    ## assuming round(pois,2)[1:floor(lambda)] > 0

    #pois.c <- pois[(floor(lambda) + 1):length(pois)]
    #rpois <- c(pois[1:max(floor(lambda),1)],pois.c[round(pois.c,2) > 0])

    lpois[[i]] <- pois
    if (length(pois) > ceiling(lambda)) {
      pois.hend <- pois[(ceiling(lambda) + 1):length(pois)]
      # Keep only ending values with prob > .005
      lpois[[i]] <- c(pois[1:ceiling(lambda)],  pois.hend[pois.hend >.005])
    }
  }

  if (nEmean == 1){
    ## lpois[[2]] id prob on rigth side
    lpois[[2]] <- lpois[[1]]
  }
  if (fill.short.side & length(lpois[[1]]) != length(lpois[[2]])){
    # make both vectors the same length
    lgth <- c(length(lpois[[1]]),length(lpois[[2]]))
    idmin <- which.min(lgth)
    zeros <- rep(0, abs(lgth[1]-lgth[2]))
    lpois[[idmin]] <- c(lpois[[idmin]],zeros)
  }

  ## for left side we reverse order of prob
  lpois[[1]] <- lpois[[1]][length(lpois[[1]]):1]

  ## if pzero is NULL, we set it to the computed value pz
  if (is.null(pzero)) pzero <- pz

  ## normalize to sum up to 1
  sumpois <- sum(c(lpois[[1]],lpois[[2]]))
  lpois[[1]] <- lpois[[1]] * (1-pzero)/sumpois
  lpois[[2]] <- lpois[[2]] * (1-pzero)/sumpois


  Pj <- c(lpois[[1]],pzero,lpois[[2]])
  attr(Pj, "lambda") <- lbda

  return(Pj)
}

#' @title
#' Distance standard errors derived from sets of MC-replicated sequences
#'
#' @description
#' Computes the mean and standard deviation of
#' each element of the pairwise distance matrix
#' across sets of MC-replicated sequences.
#'
#' @param MCrseqdata list of MC-replicated sequence datasets of class \code{stslist}. The last element is supposed to be the observed dataset.
#' @param dissrepl list, string, or object of class \code{u.diss}. If a list, list of same length as \code{MCrseqdata}. List of dissimilarity matrices or \code{dist} objects. If a character string, a method name for computing the dissimilarities with \code{\link{MCudist}}. Can also be an object of class \code{u.diss} previously computed with \code{MCudist}.
#' @param udiss logical. When \code{dissrepl} is a distance method, should distance be computed with \code{\link{MCudist}}. See details.
#' @param full.matrix logical. Should dissimilarities be organized in matrix form? Default is \code{FALSE} in which case dissimilarity matrices are converted into \code{dist} objects. If \code{TRUE}, dissimilarity \code{dist} objects are converted into matrices.
#' @param ... additional arguments passed to \code{\link{MCudist}} or \code{\link{MCdisslist}} when \code{dissrepl} is a method name.
#'
#' @details
#' Providing \code{u.diss} distances or computing distances with \code{MCudist} may be faster and can save space when the number of unique replicated sequences is smaller than the sample size times the squared root of R, which can be checked with \code{\link{MCnunique}}. When the number of unique replicated sequences largely exceeds the threshold, it is more efficient to compute distance matrices separately for each updated set of sequences with \code{\link{MCdisslist}} or by setting \code{udiss=FALSE}.
#'
#'
#' @export
#' @importFrom TraMineR seqdist
#'
#' @return Five objects:\cr
#'  \code{MCmean} Mean of distance objects over replicated sets of sequences.\cr
#'  \code{MCsd} Standard deviation of distances over replicated sets of sequences.\cr
#'  In addition, when the observed distances are provided as last element of the \code{dissrepl} list:\cr
#'  \code{MCbias} Difference between observed distance and \code{MCmean}\cr
#'  \code{MCse} Standard error of individual distances.\cr
#'  \code{MCmse} Mean square error of individual distances.\cr
#'  The five objects are of class \code{dist} when \code{attr(MCrseqdata,"toref")==FALSE} and matrices otherwise.
#'
#' @seealso \code{\link{MCseqReplicate}}, \code{\link{MCdisslist}}, \code{\link{MCudist}}, \code{\link{print.distMC}}, \code{\link{summary.distMC}}
#' @examples
#' # example code
#' exdata <- read.table(text="
#'                 a a b b
#'                 a a b b
#'                 b b a a
#'                 a c c b
#'                 b b a c
#'                 b b a c
#'                 ")
#' weights=rep(1, nrow(exdata))
#' s.exdata <- seqdef(exdata, weights = weights, id=paste("id",1:nrow(exdata), sep=""))
#'
#' ## 3 MC-replicated sequence datasets
#' altseq.list <- MCseqReplicate(s.exdata, J=1, R=3, include.obs=TRUE)
#' ## list of dissimilarity matrices
#' disslist <- MCdisslist(altseq.list, method="HAM")
#'
#' MCdselist <- MCseqdistSE(disslist)
#' print(MCdselist)
#'
#' MCratioslist <- MCratios(MCdselist)
#' print(MCratioslist)
#'

MCseqdistSE <- function(dissrepl="LCS", MCrseqdata=NULL, udiss = FALSE, full.matrix=FALSE, ...){

  if (is.null(MCrseqdata) & !inherits(dissrepl,"udist") & !is.list(dissrepl)){
    stop("MCrseqdata cannot be NULL when dissrepl is a distance method")
  }
  if (!is.list(MCrseqdata) & !is.null(MCrseqdata)){
    stop("MCrseqdata must be NULL or a list of stslist objects")
  }

  toref <- attr(dissrepl,"toref")
  if (is.null(toref)) toref <- FALSE

  if (is.character(dissrepl)){
    if (udiss){
      dissrepl <- suppressMessages(MCudist(MCrseqdata, method=dissrepl, ...))
    }
    else {
      #dissrepl <- suppressMessages(lapply(MCrseqdata, seqdist,
      #                                    method=dissrepl, full.matrix=full.matrix, ...))
      dissrepl <- MCdisslist(MCrseqdata, method=dissrepl, full.matrix=full.matrix, ...)
    }
  }
  else if (!is.list(dissrepl) & !inherits(dissrepl,"u.diss"))
    stop("Bad dissrepl type!")
  else if (!is.null(MCrseqdata)) message("MCrseqdata ignored because distances are provided!")

  attrs <- attributes(dissrepl)
  if (is.list(dissrepl) & !toref) {
    if (full.matrix)
      dissrepl <- lapply(dissrepl, as.matrix)
    else if (!full.matrix)
      dissrepl <- lapply(dissrepl, as.dist)
    attributes(dissrepl) <- attrs
  }

  ## Removing obs if present.

  diss.o <- NULL

  #print(attr(dissrepl,"obs"))

  if (is.null(attr(dissrepl,"obs"))) attr(dissrepl,"obs") <- FALSE
  if (inherits(dissrepl,"u.diss")){
    sdx <- attr(dissrepl,"sdx")
    N <- attr(dissrepl,"N")
    if (attr(dissrepl,"obs")){
      diss.o <- MCExtractDist(dissrepl, k=N)
      N <- N-1
    }
    sumDiss <- MCExtractDist(dissrepl, k=1)
    idnames <- sub("R1-","",attr(sumDiss,"Labels"))
    sumDiss2 <- sumDiss^2
    for (k in 2:N){
      sumDiss <- sumDiss + MCExtractDist(dissrepl, k=k)
      sumDiss2 <- sumDiss2 + MCExtractDist(dissrepl, k=k)^2
    }
    MCmean <- sumDiss/N
    MCsd <- sumDiss2/N - MCmean^2
  } else if (is.list(dissrepl)){
    if (attr(dissrepl,"obs")){
      diss.o <- dissrepl[[length(dissrepl)]]
      dissrepl <- dissrepl[-length(dissrepl)]
    }
    if (toref){
      idnames <- sub("R1-","",rownames(dissrepl[[1]]))
      #refnames <- paste0("ref.",colnames(dissrepl[[1]]))
    } else {
      idnames <- sub("R1-","",attr(dissrepl[[1]],"Labels"))
    }
    N <- length(dissrepl)
    MCmean <- Reduce("+", dissrepl)/N
    dissSqlist <- lapply(dissrepl, function(x){x^2})
    meanSquaredDiss <- Reduce("+", dissSqlist)/N
    MCsd <- meanSquaredDiss - MCmean^2
  }

  MCsd[MCsd<0] <- 0 ## negative values can result from numerical precision
  MCsd <- sqrt(MCsd*N/(N-1))

  attributes(dissrepl) <- attrs
  attr(dissrepl,"toref") <- toref

  if (toref){
    rownames(MCmean) <- rownames(MCsd) <-idnames
    #colnames(MCmean) <- colnames(MCsd) <- refnames
  }
  else {
    MCmean <- as.dist(MCmean)
    MCsd <- as.dist(MCsd)
    attr(MCmean,"Labels") <- attr(MCsd,"Labels") <- idnames
  }
  ret <- list(MC.mean=MCmean, MC.sd=MCsd, N=N)

  if(!is.null(diss.o)){ ## compute bias
    MCbias <- diss.o - MCmean
    MCmse <- MCsd^2 + MCbias^2
    MCse <- sqrt(MCmse)
    if (toref) {
      rownames(MCbias) <- rownames(MCmse) <- rownames(diss.o) <- idnames
      #colnames(MCbias) <- colnames(MCmse) <- colnames(diss.o) <- refnames
    }
    else {
      MCbias <- as.dist(MCbias)
      MCmse <- as.dist(MCmse)
      diss.o <- as.dist(diss.o)
      attr(MCbias,"Labels") <- attr(MCmse,"Labels") <- attr(diss.o,"Labels") <- idnames
    }
    ret <- c(ret, list(diss.o = diss.o, MC.bias=MCbias, MC.se = MCse, MC.mse=MCmse))
  }

  attr(ret,"obs") <- attr(dissrepl,"obs")
  attr(ret,"toref") <- toref
  class(ret) <- c(class(ret),list("distMC"))
  return(ret)

}

#' Ratios of distances on their standard errors
#'
#' Ratios of the observed distances to their MC standard errors and of the mean MC-simulated distances to the standard error of the mean.
#'
#' @details The standard error of the mean simulated distances is \code{mean.se = MC.se/sqrt(R)} (or \code{mean.se = MC.sd/R} when object is obtained with \code{seqdistMCSE}, because there are \code{R*R} simulated distances in that case). The ratios computed are \code{diss.z = diss.o/MC.se}, where \code{diss.o} is the distance between observed sequences, and \code{MC.mean.z = MC.mean/mean.se} with \code{MC.mean} the mean of the MC-simulated distances.
#'
#' When \code{diss.o=NULL}, the \code{diss.o} element of \code{object} is used when it exists.
#'
#' This function is handy to get afterwards ratios for outcome of \code{\link{seqdistMCSE}} obtained with \code{ratios=FALSE}.
#'
#' @param object Object of class \code{distMC} as generated by \code{\link{MCseqdistSE}} and \code{\link{seqdistMCSE}}.
#' @param diss.o Matrix or \code{dist} object. Pairwise dissimilarities between observed sequences.
#'
#' @return \code{diss.z}, \code{MC.mean.z}, and \code{mean.se} (the three as \code{dist} objects).
#'
#' @seealso \code{\link{MCseqdistSE}} and \code{\link{print.MCratios}}.
#'
#' @author Gilbert Ritschard
#' @export

MCratios <- function(object, diss.o=NULL){
  if(!inherits(object, "distMC"))
    stop("argument distMC must be a distMC object!")
  toref <- attr(object,"toref")
  if (is.null(toref)) toref <- FALSE
  newdisso <- FALSE
  MC.bias <- object[["MC.bias"]]
  if (!is.null(diss.o)){
    newdisso <- TRUE
    message("Provided diss.o used instead of the diss.o element of provided MCdist object!")
    if (!toref)
      diss.o <- as.dist(diss.o)
    MC.bias <- diss.o - object[["MC.mean"]]
  }
  else {
    if (is.null(object[["diss.o"]]))
      message("No diss.o found!")
    else
      if (toref)
        diss.o <- object[["diss.o"]]
      else
        diss.o <- as.dist(object[["diss.o"]])
  }
  R <- object[["R"]]
  N <- object[["N"]]
  if (!is.null(diss.o)){
    ## should recompute standard errors when disso was provided as diss.o argument
    if (newdisso)
      MC.se <- sqrt(object[["MC.sd"]]^2 + MC.bias^2)
    else
      MC.se <- object[["MC.se"]]
    #diss.z <- diss.o / object[["MC.sd"]]
    diss.z <- diss.o / object[["MC.se"]]
    diss.z[diss.o==0] <- 0
    diss.z[diss.z>99] <- 99 ## setting upper value o z-ratio
  } else {
    diss.z=NULL
  }
  #diss.z <- object[["diss.z"]]
  if (is.null(N))
    mean.se <- object[["MC.sd"]]/R  ## for R*R simulated distances: R = sqrt(R^2)
  else
    mean.se <- object[["MC.sd"]]/sqrt(N)
  MC.mean.z <- object[["MC.mean"]] / mean.se
  #MC.mean.z <- object[["MC.mean.z"]]

  res <- list(diss.z=diss.z, MC.mean.z = MC.mean.z, mean.se = mean.se,N=N,R=R)
  if (newdisso)
    res <- c(res, list(diss.o = diss.o, MC.se=MC.se, MC.bias=MC.bias))
  class(res) <- c(class(res),"MCratios")
  attr(res,"toref") <- toref
  return(res)
}

#############
printdistmat <- function(distmat,toref,n,nn,...){
  if (n<=0 | nn < n+1)
    print(distmat,...)
  else {
    cat("For first ",n," sequences\n")
    if(toref)
      print(distmat[1:n,], ...)
    else
      print(as.dist(as.matrix(distmat)[1:n,1:n]), ...)
    cat("...\n")
  }
}


#' Print method for distMC objects
#'
#' Prints, for each pair of the first \code{n} sequences, the mean and/or the standard deviation of the MC-replicated distances between sequences. When available, ratios are also printed by default.
#'
#' @param x \code{distMC} object as returned by \code{MCseqdistSE}.
#' @param n Integer. Number of first sequences. Default is 6. If \code{n==0} or there are less than \code{n} sequences, results are printed for all pairs of sequences.
#' @param what character string. One of \code{"mean"}, \code{"sd"}, \code{"bias"}, \code{"both"}, and \code{"all"} (default). When \code{"all"}, ratios, when present are printed together with the mean and standard deviation. When \code{"both"}, means and standard deviations are printed.
#' @param ... further arguments passed to or from other methods.
#'
#' @author Gilbert Ritschard
#' @seealso \code{\link{MCseqdistSE}}, \code{\link{summary.distMC}}.
#'
#' @return Last printed table, a \code{matrix} when \code{toref} attribute is \code{TRUE} and a \code{dist} object otherwise.
#'
#' @export
#'
#' @exportS3Method
#' print distMC

print.distMC <- function(x, n=6, what="all", ...){
  toref <- if (is.null(attr(x,"toref"))) FALSE else attr(x,"toref")
  toreftext <- " and refs,"
  if (toref){
    ns <- nrow(x[["MC.mean"]])
    toreftext <- "and refs,"
  } else {
    toreftext <- ","
    ns <- attr(x[["MC.mean"]],"Size")
  }
  N <- x[["N"]]
  R <- x[["R"]]
  RR <- if (is.null(R)) N else R*R
    cat(ns, " sequences,",RR," dissimilarity MC-replications\n")

  if (!what %in% c("all","both","mean","sd","obs","se","bias"))
    stop(' what should be one of "all","both","mean","sd","bias"')
  cat("Dissimilarity between MC-simulated sequences",toreftext,RR," dissimilarity replications")
  if (!is.null(x[["MC.sd"]])){
    if (attr(x,"obs") & what %in% c("obs","all") ){
      cat("\n diss.o: Observed dissimilarities\n")
      printdistmat(x[["diss.o"]],toref,n,ns,...)
    }
  }
  if (what %in% c("mean","both","all") ){
    cat("\n MC.mean: Mean of dissimilarities \n")
    printdistmat(x[["MC.mean"]],toref,n,ns,...)
  }
  if (!is.null(x[["MC.sd"]]) & what %in% c("sd","both","all") ){
    cat("\n MC.sd: Standard deviation of dissimilarities\n")
    printdistmat(x[["MC.sd"]],toref,n,ns,...)
  }
  if (!is.null(x[["MC.se"]]) & what %in% c("se","all") ){
    cat("\n MC.se: Standard eror of dissimilarities\n")
    printdistmat(x[["MC.se"]],toref,n,ns,...)
  }
  if (!is.null(x[["diss.z"]]) & what == "all"){
    cat("\n diss.z: Ratios diss/MC.se \n")
    printdistmat(x[["diss.z"]],toref,n,ns,...)
  }

  if (length(x[["mean.se"]]) > 0 & what == "all"){
    cat("\n MC.mean.z: Ratios MC.mean/mean.se \n")
    printdistmat(x[["MC.mean.z"]],toref,n,ns,...)

    cat("\n mean.se: Standard error of mean simulated dissimilarities \n")
    printdistmat(x[["mean.se"]],toref,n,ns,...)
  }
  if (!is.null(x[["MC.bias"]]) & what %in% c("all","bias")){
    cat("\n MC.bias: Bias (observed minus MC.mean) \n")
    printdistmat(x[["MC.bias"]],toref,n,ns,...)

#    cat("\n MC.mse: Mean squared error \n")
#    printdistmat(x[["MC.mse"]],toref,n,ns,...)
  }
}

#' Summary method for distMC objects
#'
#' Prints summary statistics of the observed dissimilarity \code{diss}, the mean \code{MC.mean}, standard deviation \code{MC.sd}, and standard error of dissimilarities between MC-replicated sequences, and the ratios  \code{diss/MC.se} and \code{MC.mean/MC.se}. Reported statistics concern all distances between original sequences.
#'
#' @param object \code{distMC} object as returned by \code{MCseqdistSE}.
#' @param ... further arguments passed to or from other methods.
#' @param silent logical: Should additional info be displayed?
#'
#' @author Gilbert Ritschard
#' @seealso \code{\link{MCseqdistSE}}, \code{\link{print.distMC}}
#'
#' @export
#'
#' @exportS3Method
#' summary distMC
#'
#' @return  \code{fivenumb} table with the statistics (min, Q1, med, Q3, max) of the observed dissimilarities, the mean, standard deviation, and standard error of the MC-simulated dissimilarities, standardized ratios, MC-bias and mean squared errors when available.



summary.distMC <- function(object, ..., silent=FALSE){
  ## need weights for stats
  toref <- attr(object,"toref")
  if (is.null(toref)) toref <- FALSE
  if (toref){
    n <- nrow(object[["MC.mean"]])
    nref <- ncol(object[["MC.mean"]])
  } else {
    n <- attr(object[["MC.mean"]],"Size")
  }
  w.u <- object[["weights"]]
  if (is.null(w.u) | length(w.u) == 1) w.u <- rep(1, n)

  #sd.w <- as.dist(w.u %*% t(w.u))
  sd.w <- as.dist(tcrossprod(w.u)) # same as above but faster

  diss.o <- object[["diss.o"]]
  R <- object[["R"]]
  N <- object[["N"]]
  #diss.z <- object[["diss.z"]]
  #MC.mean.z <- object[["MC.mean.z"]]

  if (toref){
    fivenumb <- list()
    for (i in 1:nref){
      if (any(object[["MC.mean"]] > 0) & n > 1 & !is.na(sum(w.u)) ) {
        fivenumb[[i]] <- matrix(TraMineR:::wtd.fivenum.tmr(diss.o[,i], weights=w.u), ncol=1)
        fivenumb[[i]] <- cbind(fivenumb[[i]],matrix(TraMineR:::wtd.fivenum.tmr(object[["MC.mean"]][,i], weights=w.u), ncol=1))
        cnames <- c("diss", "MC.mean")
        if (!is.null(object[["MC.sd"]])){
          fivenumb[[i]] <- cbind(fivenumb[[i]],matrix(TraMineR:::wtd.fivenum.tmr(object[["MC.sd"]][,i], weights=w.u), ncol=1))
          cnames <- c(cnames, "MC.sd")
        }
        if (!is.null(object[["mean.se"]])){
          fivenumb[[i]] <- cbind(fivenumb[[i]],matrix(TraMineR:::wtd.fivenum.tmr(object[["mean.se"]][,i], weights=w.u), ncol=1))
          fivenumb[[i]] <- cbind(fivenumb[[i]],matrix(TraMineR:::wtd.fivenum.tmr(object[["MC.mean.z"]][,i], weights=w.u), ncol=1))
          cnames <- c(cnames, "mean.se", "MC.mean/mean.se")
        }
        if (!is.null(object[["MC.bias"]])){
          fivenumb[[i]] <- cbind(fivenumb[[i]],matrix(TraMineR:::wtd.fivenum.tmr(object[["MC.se"]][,i], weights=w.u), ncol=1))
          fivenumb[[i]] <- cbind(fivenumb[[i]],matrix(TraMineR:::wtd.fivenum.tmr(object[["MC.bias"]][,i], weights=w.u), ncol=1))
          # fivenumb[[i]] <- cbind(fivenumb[[i]],matrix(TraMineR:::wtd.fivenum.tmr(object[["MC.mse"]][,i], weights=w.u), ncol=1))
          cnames <- c(cnames, "MC.se", "MC.bias")
        }
        colnames(fivenumb[[i]]) = cnames

        #rownames(fivenumb) = c("Lower notch","Lower hinge","Median","Upper hinge","Upper notch")
        rownames(fivenumb[[i]]) = c("Min","Q1","Median","Q3","Max")
      } else {
        fivenumb[[i]] <- NULL
      }
    }
    names(fivenumb) <- colnames(object[["MC.mean"]])
  } else {
    if (any(object[["MC.mean"]] > 0) & n > 1 & !is.na(sum(sd.w)) ) {
      fivenumb <- matrix(TraMineR:::wtd.fivenum.tmr(diss.o, weights=sd.w), ncol=1)
      fivenumb <- cbind(fivenumb,matrix(TraMineR:::wtd.fivenum.tmr(object[["MC.mean"]], weights=sd.w), ncol=1))
      cnames <- c("diss", "MC.mean")
      if (!is.null(object[["MC.sd"]])){
        fivenumb <- cbind(fivenumb,matrix(TraMineR:::wtd.fivenum.tmr(object[["MC.sd"]], weights=sd.w), ncol=1))
        cnames <- c(cnames, "MC.sd")
      }
      if (length(object[["mean.se"]])>0){
        fivenumb <- cbind(fivenumb,matrix(TraMineR:::wtd.fivenum.tmr(object[["mean.se"]], weights=sd.w), ncol=1))
        fivenumb <- cbind(fivenumb,matrix(TraMineR:::wtd.fivenum.tmr(object[["MC.mean.z"]], weights=sd.w), ncol=1))
        cnames <- c(cnames, "mean.se", "MC.mean/mean.se")
      }
      if (!is.null(object[["MC.bias"]])){
        fivenumb <- cbind(fivenumb,matrix(TraMineR:::wtd.fivenum.tmr(object[["MC.se"]], weights=sd.w), ncol=1))
        fivenumb <- cbind(fivenumb,matrix(TraMineR:::wtd.fivenum.tmr(object[["MC.bias"]], weights=sd.w), ncol=1))
        # fivenumb <- cbind(fivenumb,matrix(TraMineR:::wtd.fivenum.tmr(object[["MC.mse"]], weights=sd.w), ncol=1))
        cnames <- c(cnames, "MC.se", "MC.bias")
      }
      colnames(fivenumb) = cnames

      #rownames(fivenumb) = c("Lower notch","Lower hinge","Median","Upper hinge","Upper notch")
      rownames(fivenumb) = c("Min","Q1","Median","Q3","Max")
    } else {
      fivenumb <- NULL
    }
  }
  if (!silent) {
    if (is.null(N))
      cat(n," sequences, R =",R,", ", R^2, " MC-simulated dissimilarities per observed dissimilarity \n")
    else
      cat(n," sequences, N =",N," MC-simulated dissimilarities per observed dissimilarity \n")
  }

  return(fivenumb)
}


#' Print method for MCratios objects
#'
#' Prints ratios for each pair of the first \code{n} sequences.
#'
#' @param x \code{MCratios} object as returned by \code{MCratios}.
#' @param n Integer. Number of first sequences. Default is 6. If \code{n==0} or there are less than \code{n} sequences, results are printed for all pairs of sequences.
#' @param what character string. One of \code{"all"} (default), \code{"diss"}, \code{"mean"}, and \code{"se"} .
#' @param ... further arguments passed to or from other methods.
#'
#' @author Gilbert Ritschard
#' @seealso \code{\link{seqdistMCSE}}, \code{\link{MCratios}}.
#' @return Last printed table, a \code{matrix} when \code{toref} attribute is \code{TRUE} and a \code{dist} object otherwise.
#'
#' @export
#'
#' @exportS3Method
#' print MCratios

print.MCratios <- function(x, n=6, what="all", ...){

  toref <- attr(x,"toref")
  if(is.null(toref)) toref <- FALSE

  if (toref)
    nn <- nrow(x[["diss.z"]])
  else
    nn <- attr(x[["diss.z"]],"Size")
  if (!what %in% c("all","diss","mean","se"))
    stop(' what should be one of "all","diss","mean","se"')
  cat(nn, " sequences\n")
  cat("Dissimilarity between MC-simulated sequences")
  if (what %in% c("diss","all") ){
    cat("\n diss.z: Ratios diss/MC.se \n")
    printdistmat(x[["diss.z"]],toref,n,nn,...)
  }
  if (what %in% c("mean","all") ){
    cat("\n MC.mean.z: Ratios MC.mean/mean.se \n")
    printdistmat(x[["MC.mean.z"]],toref,n,nn,...)
  }
  if (what %in% c("se","all") ){
    cat("\n mean.se: Standard error of mean simulated dissimilarities \n")
    printdistmat(x[["mean.se"]],toref,n,nn)
  }
}


#' Summary method for MCratios objects
#'
#' Prints summary statistics of the ratios  \code{diss/MC.se} and \code{MC.mean/MC.se}. Reported statistics concern all distances between original sequences.
#'
#' @param object \code{MCratios} object as returned by \code{MCratios}.
#' @param weights vector of doubles. Case weights.
#' @param ... further arguments passed to or from other methods.
#' @param silent logical: Should additional info be displayed?
#' @param thresh real: threshold for counting ratios less than \code{thresh}
#'
#' @author Gilbert Ritschard
#' @seealso \code{\link{MCseqdistSE}}, \code{\link{print.distMC}}
#'
#' @export
#'
#' @exportS3Method
#' summary MCratios
#'
#' @return  \code{fivenumb} table with the statistics (min, Q1, med, Q3, max) of \code{mean.se} and the standardized ratios \code{diss.z} and \code{MC.mean.z}.

summary.MCratios <- function(object, ..., weights=NULL, silent=FALSE, thresh=2 ){
  toref <- attr(object,"toref")
  if(is.null(toref)) toref <- FALSE

  ## need weights for stats
  if (toref){
    n <-  nrow(object[["MC.mean.z"]])
    nc <- ncol(object[["MC.mean.z"]])
  } else {
    n <- attr(object[["MC.mean.z"]],"Size")
  }
  w.u <- weights
  #print(n)
  #print(w.u)
  if (is.null(w.u) | length(w.u) == 1) w.u <- rep(1, n)

  #sd.w <- as.dist(w.u %*% t(w.u))
  sd.w <- as.dist(tcrossprod(w.u)) # same as above but faster

  diss.o <- object[["diss.o"]]
  R <- object[["R"]]
  N <- object[["N"]]
  if (toref) {
    if (any(object[["mean.se"]] > 0) & n > 1 & !is.na(sum(w.u)) ) {
      fivenumb <- list()
      pLTt <- nLTt <- NULL
      ndiss <- n
      for (i in 1:nc) {
        fivenumb[[i]] <- matrix(TraMineR:::wtd.fivenum.tmr(object[["mean.se"]][,i], weights=w.u), ncol=1)
        fivenumb[[i]] <- cbind(fivenumb[[i]],matrix(TraMineR:::wtd.fivenum.tmr(object[["diss.z"]][,i], weights=w.u), ncol=1))
        fivenumb[[i]] <- cbind(fivenumb[[i]],matrix(TraMineR:::wtd.fivenum.tmr(object[["MC.mean.z"]][,i], weights=w.u), ncol=1))
        colnames(fivenumb[[i]]) = c("mean.se", "diss/MC.se", "MC.mean/mean.se")
        nnLTt <- sum(object[["diss.z"]][,i] < thresh,na.rm=TRUE)
        pLTt <- c(pLTt,nnLTt/n)
        nLTt <- rbind(nLTt,nnLTt)

        #rownames(fivenumb) = c("Lower notch","Lower hinge","Median","Upper hinge","Upper notch")
        rownames(fivenumb[[i]]) = c("Min","Q1","Median","Q3","Max")
      }
    } else {
      fivenumb <- NULL
      warning("summary not applicable!")
    }
    names(fivenumb) <- colnames(object[["mean.se"]])
  } else {
    if (any(object[["mean.se"]] > 0) & n > 1 & !is.na(sum(sd.w)) ) {
      fivenumb <- matrix(TraMineR:::wtd.fivenum.tmr(object[["mean.se"]], weights=sd.w), ncol=1)
      fivenumb <- cbind(fivenumb,matrix(TraMineR:::wtd.fivenum.tmr(object[["diss.z"]], weights=sd.w), ncol=1))
      fivenumb <- cbind(fivenumb,matrix(TraMineR:::wtd.fivenum.tmr(object[["MC.mean.z"]], weights=sd.w), ncol=1))
      colnames(fivenumb) = c("mean.se", "diss/MC.se", "MC.mean/mean.se")
      nLTt <- sum(object[["diss.z"]] < thresh,na.rm=TRUE)
      pLTt <- nLTt/(n*(n-1)/2)
      ndiss <- n*(n-1)/2
      #rownames(fivenumb) = c("Lower notch","Lower hinge","Median","Upper hinge","Upper notch")
      rownames(fivenumb) = c("Min","Q1","Median","Q3","Max")
    } else {
      fivenumb <- NULL
      warning("summary not applicable!")
    }
  }
  if (!silent) {
    if (is.null(N))
      message(n," sequences, R = ",R,", ", R^2, " MC-simulated dissimilarities per observed dissimilarity")
    else
      message(n," sequences, N = ",N," MC-simulated dissimilarities per observed dissimilarity")
  }

  return(list(fivenumb=fivenumb,prop.LT.thresh=pLTt,n.LT.thresh=nLTt,ndiss=ndiss,thresh=thresh))
}

## build table of i and j corresponding to position in dist vector

getijtable <- function(k){
  x <- NULL
  y <- NULL
  for (i in 1:(k-1)){
    x <- c(x,rep(i,k-i))
    y <- c(y,(i+1):k)
  }
  cbind(x,y)
}


## position in dist object given i and j
didx <- function(i,j,k){
  k*(i-1) - i*(i-1)/2 + j-i
}

MCdistrib <- function(dss,dur,id2=1:2, Jprob, R=100, model="keep.dss",
                      jfixed=FALSE, kchanges=NULL,
                      method="LCS", ...){
  Jprob <- Jprob/sum(Jprob)
  if (model=="keep.dss")
    ch.meth <- 1
  else if (model=="indep")
    ch.meth <- 2
  else if (model=="relative")
    ch.meth <- 3
  else
    stop("Unkown model value")

  #seq2sets <- seqdistMCSE:::seqMCset(dss[id2,],dur[id2,], Jprob=Jprob, R=R)
  #MC.dist <- seqdistMCSE:::seqdistsple(seq2sets, method=method, what="MC distances", ...)
  seq2sets <- seqMCset(dss[id2,],dur[id2,], Jprob=Jprob, R=R, ch.meth=ch.meth,
                       jfixed=jfixed, kchanges=kchanges)
  MC.dist <- seqdistsple(seq2sets, method=method, what="MC distances", ...)
  return(MC.dist)
}

# wrank adapted from wCorr package
w.rank <- function(x, w=rep(1,length(x))) {
  # sort by x so we can just traverse once
  ord <- order(x)
  ##rord <- (1:length(x))[order(ord)] # reverse order
  rord <- order(ord) # reverse order
  xp <- x[ord] # x, permuted
  wp <- w[ord] # weights, permuted
  rnk <- rep(NA, length(x)) # blank ranks vector
  # setup first itteration
  t1 <- 0 # total weight of lower ranked elements
  i <- 1 # index
  t2 <- 0 # total weight of tied elements (including self)
  n <- 0 # number of tied elements
  while(i < length(x)) {
    t2 <- t2 + wp[i] # tied weight increases by this unit
    n <- n + 1
    if(xp[i+1] != xp[i]) { # the next one is not a tie
      # find the rank of all tied elements
      rnki <- t1 + (1 + (t2-1)/2)
      # push that rank to all tied units
      for(ii in 1:n) {
        rnk[i-ii+1] <- rnki
      }
      # reset for next iteration
      t1 <- t1 + t2 # new total weight for lower values
      t2 <- 0 # new tied weight starts at 0
      n <- 0
    }
    i <- i + 1
  }
  # final row
  t2 <- t2 + wp[i] # add final weight to tied weight
  rnki <- t1 + (1 + (t2-1)/2) # final rank
  # push that rank to all final tied units
  for(ii in 1:n) {
    rnk[i-ii+1] <- rnki
  }
  # order by incoming index, so put in the original order
  rnk[rord]
}

#' @title Dissimilarities between unique replicated sequences
#'
#' @description
#' Returns the dissimilarity matrix (or \code{dist} object) between merged replicated sequences with the disaggregation indexes as attribute.
#'
#' @param MCrseqdata list of replicated \code{stslist} state sequence datasets (all of same size and with same alphabet)
#' @param method string. Name of distance method (see \code{seqdist}).
#' @param seqref state sequence object of class \code{stslist}. Fixed reference sequences.
#' @param ... Further arguments passed to \code{seqdist}
#'
#' @export
#' @importFrom WeightedCluster wcAggregateCases
#'
#' @returns object of class \code{u.diss} (pairwise dissimilarities between unique sequences) with two attributes: \code{sdx}, inverted aggregation indexes, \code{N}, number of datasets, and \code{obs}, logical indicating whether \code{k=N} corresponds to observed sequences.
#' @seealso \code{\link{MCExtractDist}}
#' @export
#' @examples
#' ## mini test data, 6 sequences of length 4, 4 unique sequences
#' exdata <- read.table(text="
#'                 a a b b
#'                 a a b b
#'                 b b a a
#'                 a c c b
#'                 b b a c
#'                 b b a c
#'                 ")
#' weights=rep(1, nrow(exdata))
#' s.exdata <- seqdef(exdata, weights = weights, id=paste("id",1:nrow(exdata), sep=""))
#'
#' ## 3 altered sequence datasets
#' (altseq.list <- MCseqReplicate(s.exdata, J=1, R=3))
#'
#' MCnunique(altseq.list, check=TRUE)
#'
#'
#' u.diss <- MCudist(altseq.list, method="LCS", full.matrix=FALSE)
#' ## Dissimilarities within first MC-set
#' MCExtractDist(u.diss, 1)
#'
#' ## list of dissimilarity matrices
#' disslist <- MCdisslist(altseq.list, use.udiss=TRUE)
#'
MCudist <- function(MCrseqdata, method="LCS", seqref=NULL, ...){

  N <- length(MCrseqdata)
  # collecting list of replicated sets into a single object
  MCrseqdata <- do.call(rbind, MCrseqdata)

  aggCases <- wcAggregateCases(MCrseqdata, weights=attr(MCrseqdata,"weights"))
  ## u.seqdata is set of unique sequences
  u.seqdata <- MCrseqdata[aggCases[["aggIndex"]],]
  sdx <- aggCases[["disaggIndex"]]
  names(sdx) <- rownames(MCrseqdata)

  ## matrices of dissimilarities between unique sequences
  if (is.null(seqref)){
    u.diss <- #suppressMessages(
      seqdist(u.seqdata, method=method, ...)
    #)
    attr(u.diss,"toref") <- FALSE
  } else { ## dist to ref
    u.diss <- seqdistToRef(u.seqdata, ref=seqref, method=method, ... )
    attr(u.diss,"toref") <- TRUE
  }

  attr(u.diss,"sdx") <- sdx
  attr(u.diss,"N") <- N
  attr(u.diss,"obs") <- attr(MCrseqdata,"obs")
  class(u.diss) <- c(class(u.diss),"u.diss")

  return(u.diss)
}

#' @title Extract k-th dissimilarity matrix from u.diss
#'
#' @param u.diss \code{u.diss} object returned by \code{MCudist}: dissimilarities between unique replicated sequences.
#' @param k integer. Subset index number for which the dissimilarity matrix must be extracted
#' @param full.matrix logical. If \code{FALSE}, the distance matrix is returned as a \code{dist} object. Ignored for distances to reference sequences.
#' @returns a dissimilarity matrix or distance object.
#' @seealso \code{\link{MCudist}}
#'
#' @export

MCExtractDist <- function(u.diss, k, full.matrix=FALSE){
  toref <- attr(u.diss,"toref")
  if (is.null(toref)) toref <- FALSE
  # length of seqdata
  sdx <- attr(u.diss,"sdx")
  N <- attr(u.diss,"N")
  n <- length(sdx)/N
  if (floor(n)*N != length(sdx))
    stop("length(sdx) must be a multiple of N")
  idk <-  (1:n)+((k-1)*n)
  if (toref){
    diss <- u.diss[sdx[idk],,drop=FALSE]
    rownames(diss) <- names(sdx[idk])
  } else {
    u.diss <- as.matrix(u.diss)
    diss <- u.diss[sdx[idk],sdx[idk],drop=FALSE]
    dimnames(diss) <- list(names(sdx[idk]),names(sdx[idk]))
  }
  if (!full.matrix & !toref){
    diss <- as.dist(diss)
  }
  return(diss)
}

#' @title Number of unique replicated sequences
#'
#' @param MCrseqdata list of replicated \code{stslist} state sequence datasets (all of same size and with same alphabet.
#' @param check logical. When \code{TRUE}, check if the number of unique replicated sequences is less than  \code{sqrt(number of replicated datasets) * (sample size)}?
#'
#'
#' @return \code{nu} number of unique replicated sequences and, when \code{check=TRUE}, \code{u.ok} the check result.
#' @importFrom WeightedCluster wcAggregateCases
#' @export
#' @seealso \code{\link{MCudist}}, \code{\link{MCseqdistSE}}
#'
MCnunique <- function(MCrseqdata, check=FALSE){
  N <- length(MCrseqdata)
  n <- nrow(MCrseqdata[[1]])

  # collecting list of replicated sets into a single object
  MCrseqdata <- do.call(rbind, MCrseqdata)
  aggCases <- wcAggregateCases(MCrseqdata, weights=attr(MCrseqdata,"weights"))
  ## u.seqdata is set of unique sequences
  u.seqdata <- MCrseqdata[aggCases[["aggIndex"]],]
  nu <- nrow(u.seqdata)
  if (check){
    u.ok <- (nu < n*sqrt(N))
    return(list(nu=nu, u.ok=u.ok))
  } else return(nu)
}

#' @title Cluster quality measures by MC-sets
#'
#' @description
#' Cluster quality measures for a range of number of groups by MC-replicated set.
#'
#'
#' @param disslist List of MC-dissimilarity matrices (or \code{dist} objects).
## #' @param diss.o
#' @param ncluster integer vector. Range of number of groups. Default is \code{2:10}.
#' @param clustmeth character. Clustering method. Either \code{"PAM"} (default) or a \code{stats::\link{hclust}} method.
#' @param weights vector of doubles. Case weights. If \code{NULL} (default), equal weights are used.
#' @param core Integer. Number of cores for parallel computing.
#' @param snow Logical. If \code{TRUE}, \code{doSNOW} is used for parallel computing, otherwise \code{doParallel} is used.
#' @param silent Logical. Should waiting and timing messages be hidden?
#' @param ... Further arguments passed to clustering functions.
#'
## #' @param what String. One of the indexes computed by \code{aricode} package such as {\code{"ARI"} (adjusted Rand index, default), \code{"RI"} (Rand index), \code{"VI"} (variation of information), \code{"NVI"}. Can also be "all \code{"all"}.
#'
#' @details
#' When \code{attr(MCdisslist,"obs")} is \code{TRUE}, the last element of \code{disslist} is treated as the dissimilarity matrix of the observed sequences.
#'
#' @return A list with two lists: \code{qual.tab}, list of tables of cluster quality statistics per MC-dissimilarity matrix, and \code{qual.max} list of cluster number $k$ for which the statistics reach their maximum (minimum for HC), \code{max.freq}, the frequency table of maximum over the MC-replicated sets, and \code{qual.obs}, cluster quality indexes for the observed sequences.
#'
#' @seealso \code{\link[WeightedCluster]{as.clustrange}},  \code{\link[WeightedCluster]{wcKMedRange}}
#'
#'
#' @importFrom WeightedCluster wcKMedRange as.clustrange
#' @importFrom doParallel registerDoParallel
#' @importFrom parallel makePSOCKcluster stopCluster makeCluster detectCores
#' @importFrom doSNOW registerDoSNOW
#' @importFrom foreach foreach %dopar% %do% %:%
#' @importFrom utils txtProgressBar setTxtProgressBar
#' @importFrom stats hclust
#'
#' @export
#'
#' @examples
#' ## mini test data, 6 sequences of length 4, 4 unique sequences
#' exdata <- read.table(text="
#'                 a a b b
#'                 a a b b
#'                 b b a a
#'                 a c c b
#'                 b b a c
#'                 b b a c
#'                 ")
#' weights=rep(1, nrow(exdata))
#' s.exdata <- seqdef(exdata, weights = weights, id=paste("id",1:nrow(exdata), sep=""))
#'
#' ## 3 altered sequence datasets
#' set.seed(25)
#' altseq.list <- MCseqReplicate(s.exdata, J=1, R=3)
#' ## list of dissimilarity matrices
#' disslist <- MCdisslist(altseq.list, method="LCS")
#' diss.o <- seqdist(s.exdata, method="LCS")
#' ## cluster per MC-dissimilarity matrices
#' res <- MCclustqual(disslist,ncluster=3)
#' res
#'
#'
MCclustqual <- function(disslist, ncluster=10, clustmeth="PAM",  weights=NULL,
                        core=1, snow=TRUE, silent=FALSE, ...){

  N <- length(disslist)
  isobs <- attr(disslist,"obs")
  if (is.null(isobs)) isobs <- FALSE

  if (inherits(disslist[[1]],"dist")){
    ncases <- attr(disslist[[1]],"Size")
  } else {
    ncases <- nrow(disslist[[1]])
  }
  if (is.null(weights)) {
    weights <- rep(1, ncases)
    weighted=FALSE
  } else {
    if (length(weights) != ncases)
      stop("length of weights not equal to number of cases!")
    if (all(weights==1))
      weighted <- FALSE
    else weighted <- TRUE
  }

  if (!silent & (core==1 | snow)) {
    #cat("\n")
    progress <- function(n) setTxtProgressBar(pb, n)
    opts <- list(progress = progress)
    pb <- txtProgressBar(min=1, max=N, initial=1, style=3)
  }
  else opts <- list()

  KMedRangeStats <- function(diss, kvals, weights, ...){
    wcKMedRange(diss, kvals=kvals, weights=weights, ...)$stats
  }
  hclustRange <- function(diss, clustmeth, ncluster, weights, ... ){
    hcl <- hclust(as.dist(diss), method=clustmeth, members=weights)
    as.clustrange(hcl, diss=diss, ncluster=ncluster, weights=weights, ...)$stats
  }
  if (core==1) {
    if (clustmeth %in% c("PAM","KM")){
      qlist <- lapply(disslist, KMedRangeStats, kvals=2:ncluster, weights=weights, ...)
    } else {
      qlist <- lapply(disslist, hclustRange, clustmeth=clustmeth, ncluster=ncluster, weights=weights, ...)
    }
    if (!silent) close(pb)
  } else { ## parallel computing
    cl <- makeCluster(core, type="SOCK")
    if (snow)
      registerDoSNOW(cl)
    else
      registerDoParallel(cl)

    qlist <- foreach (i = 1:N, .combine=c, .packages=c("WeightedCluster"), .options.snow=opts) %dopar% {
      if (clustmeth %in% c("PAM","KM")){
        list(KMedRangeStats(disslist[[i]], kvals=2:ncluster, weights=weights, ...))
      } else {
        list(hclustRange(disslist[[i]], clustmeth=clustmeth, ncluster=ncluster, weights=weights, ...))
      }

    }
    stopCluster(cl)
    if(!silent & snow) {
      close(pb)
    }
  }

  if (isobs){
    qobs <- qlist[[N]]
    qlist <- qlist[-N]
    N <- N-1
  } else {
    qobs <- NULL
  }


  which.max.vect <- function(m) {
    stats <- m
    stats[,"HC"] <- -stats[,"HC"] ## For HC we seek the min
    apply(stats,2,which.max) + 1
    }
  max.list <- lapply(qlist, which.max.vect)

  ## distribution of max
  tabmax <- do.call(rbind,max.list)
  tabmaxf <- data.frame(factor(tabmax[,1],levels=2:ncluster))
  #print(tabmaxf)
  maxdist <- data.frame(table(tabmaxf[,1]))
  #print(maxdist)
  rownamesd <- maxdist[,1]
  maxdist <- maxdist[,2]
  for (i in 2:ncol(tabmax)){
    tabmaxf <- cbind(tabmaxf,factor(tabmax[,i],levels=2:ncluster))
    maxdist <- cbind(maxdist,data.frame(table(tabmaxf[,i]))[,2])
  }
  rownames(maxdist) <- rownamesd
  colnames(tabmaxf) <- colnames(maxdist) <- colnames(tabmax)

  ret <- list(qual.tab = qlist, qual.max = tabmaxf, max.freq=maxdist, qual.obs=qobs)

  class(ret) <- c(class(ret),"MCclustQ")
  attr(ret,"obs") <- attr(disslist,"obs")
  return(ret)
}

#' @title Correlation between observed and MC-simulated distances
#'
#' @param disslist List of matrices or dist objects: the MC-replicated dissimilarities
#' @param diss.o Matrix or dist object: Observed dissimilarities
#' @param method String. One of \code{"Spearman"} (default) and \code{"Pearson"}.
#' @param weights vector of doubles. Case weights. If \code{NULL} (default), equal weights are used.
## #' @param what String. One of \code{"corr"} (correlations, default).
## #' @param core Integer. Number of cores for parallel computing.
## #' @param snow Logical. If \code{TRUE}, \code{doSNOW} is used for parallel computing, otherwise \code{doParallel} is used.
## #' @param silent Logical. Should waiting and timing messages be hidden?
#'
#' @details
#' When \code{diss.o=NULL}, the last element of \code{disslist} is taken as \code{diss.o} and the other elements as sets of MC-replicated dissimilarities.
#'
#' @return vector of correlation between observed and MC-dissimilarities.
#'
#'
#' @importFrom wCorr  weightedCorr
#' @importFrom utils txtProgressBar setTxtProgressBar
#'
#' @export
#'
#' @examples
#' ## mini test data, 6 sequences of length 4, 4 unique sequences
#' exdata <- read.table(text="
#'                 a a b b
#'                 a a b b
#'                 b b a a
#'                 a c c b
#'                 b b a c
#'                 b b a c
#'                 ")
#' weights=rep(1, nrow(exdata))
#' s.exdata <- seqdef(exdata, weights = weights, id=paste("id",1:nrow(exdata), sep=""))
#'
#' ## 3 altered sequence datasets
#' set.seed(25)
#' altseq.list <- MCseqReplicate(s.exdata, J=1, R=3, include.obs=TRUE)
#' ## list of dissimilarity matrices
#' disslist <- MCdisslist(altseq.list)
#' MCdisscorr(disslist)
#'
#'
MCdisscorr <- function(disslist, diss.o=NULL, method="Spearman", weights=NULL){

  toref <- attr(disslist,"toref")
  if (is.null(toref)) toref <- FALSE

  N <- length(disslist)
  if (is.null(diss.o)){
    message("Extracting diss.o from disslist")
    diss.o <- disslist[[N]]
    disslist <- disslist[-N]
    N <- N-1
  }

  ## setting weights
  if (inherits(diss.o,'dist')){
    ncases <- attr(diss.o,'Size')
  } else {
    ncases <- nrow(diss.o)
  }
  if (is.null(weights)) {
    weights <- rep(1, ncases)
    weighted=FALSE
  } else {
    if (length(weights) != ncases)
      stop("length of weights not equal to number of cases!")
    if (all(weights==1))
      weighted <- FALSE
    else weighted <- TRUE
  }

  if (toref){
    nc <- ncol(diss.o)
    corr <- list()
    for (i in 1:nc){
      corr[[i]] <- sapply(disslist, function(x){wCorr::weightedCorr(x=x[,i], y=diss.o[,i], method=method, weights=weights)})
    }
    names(corr) <- colnames(diss.o)

  } else {
    w <- as.dist(tcrossprod(weights))
    corr <- sapply(disslist, function(x){wCorr::weightedCorr(x=as.dist(x), y=as.dist(diss.o), method=method, weights=w)})
  }

  return(corr)
}

## timing change methods

## definition of move depends on Jprob
move <- function(n, Jprob){
  #print(n)
  #print(Jprob)
  if (length(Jprob) == 1){
    #JJ <- Jprob * 2 + 1
    return(floor(runif(n) * (Jprob * 2 + 1) - Jprob))
  } else {
    Jprob <- Jprob/sum(Jprob) ## already normalized
    JJ <- length(Jprob)
    Jdist <- cumsum(Jprob)
    ## drop intervals with zero prob
    nonzero <- which(Jprob != 0)
    Jprob <- Jprob[nonzero]
    Jdist <- Jdist[nonzero]
    JL <- length(Jdist)
    p <- runif(n)
    #print(p)
    res <- sapply(p, function(x) {which(Jdist >= x)[1]})
    #print(res)
    #print(nonzero)
    res[res<=0] <- 1
    res[res > JL] <- JL ## should never happen
    res <- nonzero[res] - (JJ + 1)/2
    #print(res)
    return(res)
  }
}

setklist <- function(k, nt){
  if (is.null(k))
    k <- nt
  else if (k <0) {
    k <- floor(runif(1,min=0,max=nt)+.5)
  }
  if (k < nt)
    klist <- sample(nt,k)
  else
    klist <- 1:nt
  return(klist)
}

## method="keep.dss"
ch.dur <- function(sduri, Jprob, jfixed=TRUE, k=NULL){
  #J <- length(Jprob)
  #JJ <- 2*J+1
  nt <- sum(!is.na(sduri))-1

  ##change <- floor(runif(nt) * JJ - J)
  if (nt > 0) {
    klist <- setklist(k, nt)
    change <- integer(nt)
    if (jfixed)
      change[klist] <- move(1, Jprob=Jprob)
    else
      change[klist] <- move(length(klist), Jprob=Jprob)
    #print(change)
    for (i in klist) {
      if (change[i] < 0)
        change[i] <- -min(-change[i], sduri[i] - 1)
      else if (change[i] >0)
        change[i] <- min(change[i], sduri[i+1] - 1)

      sduri[i] <- sduri[i] + change[i]
      sduri[i+1] <- sduri[i+1] - change[i]
    }
  }
  return(sduri)
}


## method="indep"
ch.dur.indep <- function(sduri, Jprob, jfixed=FALSE, k=NULL){
  #J <- length(Jprob)
  #JJ <- 2*J+1
  nt <- sum(!is.na(sduri))-1

  ##change <- floor(runif(nt) * JJ - J)
  if (nt > 0) {
    klist <- setklist(k, nt)
    change <- integer(nt)
    if (jfixed)
      change[klist] <- move(1, Jprob=Jprob)
    else
      change[klist] <- move(length(klist), Jprob=Jprob)
    #print(change)
    for (i in klist) {
      if (change[i] < 0){
        chg <- -change[i]
        sduri[i+1] <- sduri[i+1] + chg
        if (sduri[i] >= chg){
          sduri[i] <- sduri[i] - chg
        } else {
          rchg <- chg - sduri[i]
          sduri[i] <-0
          ii <- i-1
          while (ii > 0 & rchg > 0){
            dur <- sduri[ii]
            sduri[ii] <- max(0, dur - rchg)
            rchg <- rchg - (dur - sduri[ii])
            ii <- ii-1
          }
          if (rchg > 0) { ## must adjust change at i+1
            sduri[i+1] <- sduri[i+1] - rchg
          }
        }
      }
      else {
        chg <- change[i]
        sduri[i] <- sduri[i] + chg
        if (sduri[i+1] >= chg){
          sduri[i+1] <- sduri[i+1] - chg
        } else {
          rchg <- chg - sduri[i+1]
          sduri[i+1] <-0
          ii <- i+2
          while (ii < (nt + 1) & rchg > 0){
            dur <- sduri[ii]
            sduri[ii] <- max(0, dur - rchg)
            rchg <- rchg - (dur - sduri[ii])
            ii <- ii+1
          }
          if (rchg > 0) { ## must adjust change at i
            sduri[i] <- sduri[i] - rchg
          }
        }
      }
    }
  }
  return(sduri)
}

## method="relative"
ch.dur.relat <- function(sduri, Jprob, jfixed=FALSE, k=NULL){
  #J <- length(Jprob)
  #JJ <- 2*J+1
  nt <- sum(!is.na(sduri))-1

  ##change <- floor(runif(nt) * JJ - J)
  if (nt > 0) {
    klist <- setklist(k, nt)
    change <- integer(nt)
    if (jfixed)
      change[klist] <- move(1, Jprob=Jprob)
    else
      change[klist] <- move(length(klist), Jprob=Jprob)
    #print(change)
    for (i in klist) {
      if (change[i] < 0){
        chg <- -change[i]
        sduri[nt+1] <- sduri[nt+1] + chg
        if (sduri[i] >= chg){
          sduri[i] <- sduri[i] - chg
        } else {
          rchg <- chg - sduri[i]
          sduri[i] <-0
          ii <- i-1
          while (ii > 0 & rchg > 0){
            dur <- sduri[ii]
            sduri[ii] <- max(0, dur - rchg)
            rchg <- rchg - (dur - sduri[ii])
            ii <- ii-1
          }
          if (rchg > 0) { ## must adjust change at nt+1 (last spell)
            sduri[nt+1] <- sduri[nt+1] - rchg
          }
        }
      }
      else {
        chg <- change[i]
        sduri[i] <- sduri[i] + chg
        if (sduri[nt+1] >= chg){
          sduri[nt+1] <- sduri[nt+1] - chg
        } else {
          rchg <- chg - sduri[nt+1]
          sduri[nt+1] <-0
          ii <- nt
          while (ii > i & rchg > 0){
            dur <- sduri[ii]
            sduri[ii] <- max(0, dur - rchg)
            rchg <- rchg - (dur - sduri[ii])
            ii <- ii-1
          }
          if (rchg > 0) { ## must adjust change at i
            sduri[i] <- sduri[i] - rchg
          }
        }
      }
    }
  }
  return(sduri)
}

#' @title Correlation between 1st MDS factor of observed and MC-simulated distances
#'
#' @param disslist List of matrices or dist objects: the MC-replicated dissimilarities
#' @param diss.o Matrix or dist object: Observed dissimilarities
#' @param method String. One of \code{"Spearman"} (default) and \code{"Pearson"}.
#' @param weights vector of doubles. Case weights. If \code{NULL} (default), equal weights are used.
#' @param what String. One of \code{"corr"} (correlations, default), \code{"mds"} (list of mds scores), and \code{"both"}.
#' @param core Integer. Number of cores for parallel computing.
#' @param snow Logical. If \code{TRUE}, \code{doSNOW} is used for parallel computing, otherwise \code{doParallel} is used.
#' @param silent Logical. Should waiting and timing messages be hidden?
#'
#' @details
#' When \code{diss.o=NULL}, the last element of \code{disslist} is taken as \code{diss.o} and the other elements as sets of MC-replicated dissimilarities.
#'
#' @return when \code{what="corr"}, vector of correlation between mds of dissimilarities in MC-replicated sets, when \code{what="mds"}, of first mds scores, and when \code{what="both"}, list with \code{corr} as first element and \code{mdslist}, the list of mds scores as second element.
#'
#'
#' @importFrom vegan  wcmdscale
#' @importFrom wCorr  weightedCorr
#' @importFrom doParallel registerDoParallel
#' @importFrom parallel makePSOCKcluster stopCluster makeCluster detectCores
#' @importFrom doSNOW registerDoSNOW
#' @importFrom foreach foreach %dopar% %do% %:%
#' @importFrom utils txtProgressBar setTxtProgressBar
#'
#' @export
#'
#' @examples
#' ## mini test data, 6 sequences of length 4, 4 unique sequences
#' exdata <- read.table(text="
#'                 a a b b
#'                 a a b b
#'                 b b a a
#'                 a c c b
#'                 b b a c
#'                 b b a c
#'                 ")
#' weights=rep(1, nrow(exdata))
#' s.exdata <- seqdef(exdata, weights = weights, id=paste("id",1:nrow(exdata), sep=""))
#'
#' ## 3 altered sequence datasets
#' set.seed(25)
#' altseq.list <- MCseqReplicate(s.exdata, J=1, R=3)
#' ## list of dissimilarity matrices
#' disslist <- MCdisslist(altseq.list)
#' MCmdscorr(disslist)
#'
#'
MCmdscorr <- function(disslist, diss.o=NULL, method="Spearman", weights=NULL, what="corr",
                      core=1, snow=TRUE, silent=FALSE){
  message(method," correlations between 1st MDS factor of observed and replicated distances")
  N <- length(disslist)
  if (is.null(diss.o)){
    message("Extracting diss.o from disslist")
    diss.o <- disslist[[N]]
    disslist <- disslist[-N]
    N <- N-1
  }

  ## setting weights
  if (inherits(diss.o,'dist')){
    ncases <- attr(diss.o,'Size')
  } else {
    ncases <- nrow(diss.o)
  }
  if (is.null(weights)) {
    weights <- rep(1, ncases)
    weighted=FALSE
  } else {
    if (length(weights) != ncases)
      stop("length of weights not equal to number of cases!")
    if (all(weights==1))
      weighted <- FALSE
    else weighted <- TRUE
  }

  message("Computing first MDS scores of observed dissimilarities")
  mds.o <- vegan::wcmdscale(diss.o, k=1, w=weights)
  message("Computing first MDS scores of replicated dissimilarities")
  if (!silent & (core==1 | snow)) {
    #cat("\n")
    progress <- function(n) setTxtProgressBar(pb, n)
    opts <- list(progress = progress)
    pb <- txtProgressBar(min=1, max=N, initial=1, style=3)
  }
  else opts <- list()

  if (core==1){
    #mdslist <- lapply(disslist,wcmdscale,k=1, w=weights)
    mdslist <- list()
    for (i in 1:N){
      if (!silent) setTxtProgressBar(pb, i)
      mdslist[[i]] <- vegan::wcmdscale(disslist[[i]],k=1, w=weights)
    }
    if (!silent) close(pb)
  }
  else { ## parallel computing
    #pb <- txtProgressBar(min=1, max=k-1, initial=1, style=3)
    #cl <- makePSOCKcluster(core)
    cl <- makeCluster(core, type="SOCK")
    if (snow)
      registerDoSNOW(cl)
    else
      registerDoParallel(cl)

    mdslist <- foreach (i = 1:N, .combine=c, .options.snow=opts) %dopar% {
      list(vegan::wcmdscale(disslist[[i]],k=1, w=weights))
    }

    stopCluster(cl)
    if(!silent & snow) {
      close(pb)
    }
  }


  if (what != "mds"){
    message("Computing correlations")
    corr <- sapply(mdslist, function(x){wCorr::weightedCorr(x=x, y=mds.o, method=method, weights=weights)})
    for (i in which(corr < 0)) {
      corr[i] <- -corr[i]
      mdslist[[i]] <- -mdslist[[i]]
    }
  }
  mdslist <- c(mdslist, list(mds.o=mds.o))

  if (what == "mds")
    return(mdslist)
  else if (what %in% c("cor","corr"))
    return(corr)
  else if (what == "both"){
    return(list(corr=corr, mdslist = mdslist))
  }


}

#' @title
#' Generate R altered sequence data sets.
#'
#' @description
#' R \code{stslist} state sequence objects are generated by applying the chosen timing error model to the provided state sequence object.
#'
#' @details
#' This function is handy for testing how outcome of a sequence analysis may vary with timing errors in the reported sequences.
#'
#' Use the vector form of \code{J} to specify the probability distribution of the timing error. See function \code{\link{MCpj}} to generate a probability vector that complies with expected mean timing errors.
#'
#' @param seqdata A state sequence \code{stslist} object as generated by \code{\link[TraMineR]{seqdef}}.
#' @param J Integer or vector of positive numbers. If an integer, maximal timing error (number of unit periods around first state of new spell. Default is \code{J=1}. If a vector, weights of the timing errors \code{k = -K, -(K-1), ..., K-1, K}, where \code{2K + 1} is the length of \code{J}. The vector length must be odd.
#' @param R Integer. Number of random replicated sequence data. Default is \code{R=10}. The weights are internally normalized to sum to 1.
#' @param jfixed Logical. Should same error j be applied to all transitions in a sequence? Default is \code{FALSE}.
#' @param kchanges Integer, string, or \code{NULL}. If integer, number of transitions whose time can potentially be altered in each sequence. If \code{"rand"}, the number of potential changes is randomly selected for each sequence. If \code{NULL} (default), all transitions can potentially be altered.
#' @param silent Logical. Should waiting and timing messages be hidden?
#' @param unique Logical. Should only unique sequences be replicated? Default is \code{FALSE}. If \code{TRUE} weights will reflect the multiple occurrences of each original unique sequence.
#' @param model String. Time alteration model. One of \code{"keep.dss"} (default), \code{"indep"} (suppress spells erased by move of transition), and \code{"relative"} (keep time until next transitions unchanged).
#' @param include.obs logical. Should the observed sequence data be added as last element.
#'
#' @return List of R altered \code{stslist} objects plus observed sequence object as last element when \code{include.obs=TRUE}.
#'
#' @seealso \code{\link{MCpj}}
#'
#' @references
#' Ritschard, G. and Liao, T.F. (2026). Assessing the Impact of Timing Errors in Sequence Analysis. \emph{International Journal of Social Research Methodology}. Forthcoming
#'
#'
#' @author Gilbert Ritschard
#'
#' @export
#'
## @exportS3Method print distMC
## @exportS3Method summary distMC
#'
#' @import TraMineR
#' @importFrom WeightedCluster wcAggregateCases
#'
#'
#' @examples
#' ## mini test data, 6 sequences of length 4, 4 unique sequences
#' exdata <- read.table(text="
#'                 a a b b
#'                 a a b b
#'                 b b a a
#'                 a c c b
#'                 b b a c
#'                 b b a c
#'                 ")
#' weights=rep(1, nrow(exdata))
#' s.exdata <- seqdef(exdata, weights = weights, id=paste("id",1:nrow(exdata), sep=""))
#'
#' ## 3 altered sequence datasets
#' (altseq.list <- MCseqReplicate(s.exdata, J=1, R=3))
#'
#' ## list of dissimilarity matrices
#' suppressMessages(dist.list <- lapply(altseq.list, seqdist, method="LCS", full.matrix=FALSE))
#' dist.list
#'
#' ## Can also be obtained with MCdisslist, which offers option use.udiss;
#' ## use.udiss=TRUE is faster when number of unique merged replicated
#' ## sequences is less than n*sqrt(R).
#' suppressMessages(dist.list <- MCdisslist(altseq.list, method="LCS", use.udiss=TRUE))
#'
#' ## Replication based on expected left and right non-zero errors of 1.1
#' ##  and assuming a 0.5 probability of no error
#' Pj <- MCpj(Emean=1.1, pzero=.5)
#' (altseq2.list <- MCseqReplicate(s.exdata, J=Pj, R=3))


MCseqReplicate <- function(seqdata, J=1, R=20, silent=FALSE, unique=FALSE, model="keep.dss", jfixed=FALSE, kchanges=NULL, include.obs=FALSE){

  if (model=="keep.dss"){
    ch.meth <- 1
  }else if (model=="indep"){
    ch.meth <- 2
  }else if (model=="relative"){
    ch.meth <- 3
  }else {
    stop("Unkown model value")
  }

  if (is.character(kchanges) ){
    if (kchanges=="rand"){
      kchanges <- -1
    }else
      stop("Invalid kchanges value!")
  }

  if(length(J) == 1) {
    JJ <- 2*J + 1
    Jprob <- J
  } else {
    JJ <- length(J)
    if (! JJ %% 2)
      stop("Length of J must be odd!")
    if (any(J<0))
      stop("Negative values in J")

    Jprob <- J/sum(J)
    J <- (JJ - 1)/2
  }

  if(!inherits(seqdata,"stslist"))
    stop("argument seqdata must be a stslist object!")


  ## extract unique sequences (we do not need weights here)

  if(unique) {
    aggCases <- wcAggregateCases(seqdata, weights=attr(seqdata,"weights"))
    ## u.seqdata is set of unique sequences
    u.seqdata <- seqdata[aggCases[["aggIndex"]],]
    weights <- aggCases[["aggWeights"]]
    sdx <- aggCases[["disaggIndex"]]
    names(sdx) <- rownames(seqdata)
  } else {
    aggCases <- NULL
    u.seqdata <- seqdata
    weights <- attr(seqdata,"weights")
    sdx <- 1:nrow(seqdata)
  }
  ## Transform into numeric to avoid issues with unwanted character in state labels
  #DataU <- seqasnum.one(u.seqdata) ## returns numeric for missing
  DataU <- seqnum(u.seqdata) ## returns a stslist
  #print(DataU)
  L <- ncol(DataU)  ## sequence length if all sequences are of equal length
  NU <- nrow(DataU)

  sDataU <- DataU ## DataU is stslist
  sdss <- seqdss(sDataU)
  sdur <- seqdur(sDataU)

  res <- MClistrepset(dss=sdss, sdur=sdur, Jprob=Jprob, R=R, ch.meth=ch.meth, jfixed=jfixed, kchanges=kchanges)


  nrs <- nrow(res)/R

  reslist <- list()
  alph <- alphabet(u.seqdata)
  cnames <- colnames(u.seqdata)
  labels <- attr(u.seqdata,"labels")
  for (i in 1:R){
    reslist[[i]] <- res[((i-1)*nrs + 1):(i*nrs),]
    attr(reslist[[i]],"weights") <- weights
    alphabet(reslist[[i]]) <- alph
    attr(reslist[[i]],"labels") <- labels
    colnames(reslist[[i]]) <- cnames
    #print(attr(reslist[[i]],"alphabet"))
  }
  if (include.obs){
    reslist[[R+1]] <- u.seqdata
    attr(reslist[[R+1]],"weights") <- weights
    colnames(reslist[[i]]) <- cnames
  }

  attr(reslist,"unique") <- unique
  attr(reslist,"obs") <- include.obs

  #invisible(reslist)
  reslist
}


## stslist of all replicated sequences
MClistrepset <-function(dss,sdur,
                        Jprob,R=10,ch.meth,
                        jfixed, kchanges){
  mdss <- as.matrix(dss)
  ncs <- ncol(mdss)
  nrs <- nrow(mdss)
  #setlist <- list()
  setsps <- NULL
  for (i in 1:R){
    if (ch.meth==1){
      dur <- t(apply(sdur,1,ch.dur,Jprob=Jprob, jfixed = jfixed, k=kchanges))
    }else if (ch.meth==2){
      dur <- t(apply(sdur,1,ch.dur.indep,Jprob=Jprob, jfixed = jfixed, k=kchanges))
    }else if (ch.meth==3){
      dur <- t(apply(sdur,1,ch.dur.relat,Jprob=Jprob, jfixed = jfixed, k=kchanges))
    }

    if (nrow(dur)==1) dur <- t(dur)
    wdurlist <- apply(dur,1, function(x){which(x>0)}, simplify=FALSE)
    set <- matrix(nrow=nrs,ncol=ncs)
    for (j in 1:nrs){
      set[j,1:length(wdurlist[[j]])] <- paste0(mdss[j,wdurlist[[j]]],"/",dur[j,wdurlist[[j]]])
    }
    rownames(set) <- paste0("R",i,"-",rownames(dss))
    setsps <- rbind(setsps,set)
  }
  setsts <- seqformat(setsps, from="SPS", to="STS", SPS.in=list(xfix="", sdsep="/"), stsep="-")
  rownames(setsts) <- rownames(setsps)
  seqset <- suppressMessages(
      seqdef(setsts, alphabet=alphabet(dss))
      )
  seqset
}

#' @title Comparing MC-clusters with cluster of observed data
#'
#' @description
#' Comparison indexes between clusters based on observed data and each of MC-replicated clusters.
#'
#'
#' @param clustlist List of MC-replicated vectors of cluster memberships.
#' @param clust.o Cluster memberships based on observed dissimilarities.
#' @param weights vector of doubles. Case weights. If \code{NULL} (default), equal weights are used.
## #' @param what String. One of the indexes computed by \code{aricode} package such as {\code{"ARI"} (adjusted Rand index, default), \code{"RI"} (Rand index), \code{"VI"} (variation of information), \code{"NVI"}. Can also be "all \code{"all"}.
#'
#' @details
#' When \code{diss.o=NULL}, the last element of \code{disslist} is taken as \code{diss.o} and the other elements as sets of MC-replicated dissimilarities.
#'
#' @return A table with in columns the list of comparison scores provided by \code{aricode::\link[aricode]{clustComp}} for each replicated set, except Chi2, which is replaced by Cramer's V.
#'
#' @seealso \code{\link[aricode]{clustComp}}
#'
#'
## #' @importFrom mclust adjustedRandIndex
#' @importFrom aricode clustComp
#'
#' @export
#'
#' @examples
#' ## mini test data, 6 sequences of length 4, 4 unique sequences
#' exdata <- read.table(text="
#'                 a a b b
#'                 a a b b
#'                 b b a a
#'                 a c c b
#'                 b b a c
#'                 b b a c
#'                 ")
#' weights=rep(1, nrow(exdata))
#' s.exdata <- seqdef(exdata, weights = weights, id=paste("id",1:nrow(exdata), sep=""))
#'
#' ## 3 altered sequence datasets
#' set.seed(25)
#' altseq.list <- MCseqReplicate(s.exdata, J=1, R=3)
#' ## list of dissimilarity matrices
#' disslist <- MCdisslist(altseq.list, method="LCS")
#' diss.o <- seqdist(s.exdata, method="LCS")
#' ## cluster per MC-dissimilarity matrices
#' library(WeightedCluster)
#' clust.o <- wcKMedoids(diss.o, k=2, cluster.only=TRUE)
#' clustlist <- lapply(disslist, wcKMedoids, k=2, cluster.only=TRUE)
#' res <- MCclustcomp(clustlist, clust.o=clust.o)
#' res
#'
#'
MCclustcomp <- function(clustlist, clust.o=NULL, weights=NULL){
  message(" Cluster comparison scores between MC-Clusters and clusters of observed data")

  N <- length(clustlist)
  if (is.null(clust.o)){
    message("Extracting clust.o from clustlist")
    clust.o <- clustlist[[N]]
    clustlist <- clustlist[-N]
    N <- N-1
  }
  n <- length(clust.o)
  k <- length(unique(clust.o))

  if (!is.null(weights))
    message("Weights currently not supported! Weights are ignored.")
    ## setting weights
  # ncases <- length(clust.o)
  # if (is.null(weights)) {
  #   weights <- rep(1, ncases)
  #   weighted=FALSE
  # } else {
  #   if (length(weights) != ncases)
  #     stop("length of weights not equal to number of cases!")
  #   if (all(weights==1))
  #     weighted <- FALSE
  #   else weighted <- TRUE
  # }

  ## should write own function for weighted data
  ## aricode::clustComp does not support weights

  ret <- lapply(clustlist, function(x) {as.matrix(unlist(aricode::clustComp(x,clust.o)))})
  ret <- do.call(cbind,ret)
  ret["Chi2",] <- sqrt(ret["Chi2",]/(n*(k-1)))
  rownames(ret)[rownames(ret)=="Chi2"] <- "V"
  ret
}

#' @title List of dissimilarity matrices
#'
#' @description
#'  Compute the dissimilarity matrix for each of the provided sets of sequences.
#'
#' @param MCrseqdata List of state sequence objects of class \code{stslist}.
#' @param method string. Name of a distance method (see \code{\link[TraMineR]{seqdist}}).
#' @param seqref state sequence object of class \code{stslist}. Fixed reference sequences.
#' @param use.udiss logical. Should computation be based on unique sequences?
#' @param full.matrix logical. Should pairwise distances be returned in matrix form? If \code{FALSE} (default),
#' a list of \code{dist} objects is returned. Applies only when \code{seqref=NULL}.
#' @param ... further arguments passed to \code{seqdist}.
#'
#' @details
#' When \code{use.udiss=TRUE}, the function first computes dissimilarities between unique merged replicated sequences through a single call to \code{seqdist()} and the set of dissimilarity matrices are then extracted from the resulting distance matrix. This is generally faster when the number of unique merged replicated sequences is less than \code{sqrt(number of replicated datasets) * (sample size)}, which can be checked with \code{\link{MCnunique}}.
#'
#' @return list of dissimilarity matrices or \code{dist} objects with logical attribute \code{"obs"}, which is \code{TRUE} when the list includes the dissimilarities between observed sequences as last element.
#' @seealso \code{\link{MCseqReplicate}}, \code{\link{MCudist}} and examples in their help pages.
#' @export

MCdisslist <- function(MCrseqdata, method="LCS", seqref=NULL, full.matrix=FALSE, use.udiss=FALSE, ...){
  if (is.null(seqref)){
    if (use.udiss){
      udiss <- suppressMessages(MCudist(MCrseqdata, method=method, full.matrix=full.matrix, ...))
      disslist <- list()
      for (k in 1:length(MCrseqdata)){
        disslist[[k]] <- MCExtractDist(udiss,k,full.matrix=full.matrix)
      }
    } else {
      disslist <- suppressMessages(lapply(MCrseqdata, seqdist, method=method,
                                        full.matrix=full.matrix, ...))
    }
    attr(disslist,"toref") <- FALSE
  } else { ## distances to fixed reference sequences
    if (use.udiss){
      udiss <- suppressMessages(MCudist(MCrseqdata, method=method, seqref=seqref, ...))
      disslist <- list()
      for (k in 1:length(MCrseqdata)){
        disslist[[k]] <- MCExtractDist(udiss,k)
      }

    } else {
    disslist <- suppressMessages(lapply(MCrseqdata, seqdistToRef,
                                        ref = seqref,
                                        method=method,
                                        ...))
    }
    attr(disslist,"toref") <- TRUE
  }
  attr(disslist,"obs") <- attr(MCrseqdata,"obs")
  return(disslist)
}

seqdistToRef <- function(s, ref, method, ...){
  nseq <- nrow(s)
  nref <- nrow(ref)
  rownames(ref) <- paste0("ref.",rownames(ref))
  seqr <- rbind(s,ref)
  dist <- seqdist(seqr, refseq = list(1:nseq,nseq+(1:nref)), method=method, ...)
}

#' @title
#' Mean and standard deviation of dissimilarities between pairs of randomly altered sequences.
#'
#' @description
#' For each pair of sequences, returns the mean and standard deviation (MCSE) of the dissimilarities between all combinations of MC-replicated sequences, where sequences are replicated with random timing changes.
#'
#' @details
#' Let \eqn{B_x} be the set of \code{R} sequences derived from a sequence \eqn{x} by randomly altering the timing of the transitions (state changes) in \eqn{x}. The MC standard error of the dissimilarity \eqn{d(x,y)} between two sequences \eqn{x} and \eqn{y} is the empirical standard deviation of the dissimilarities between the sequences of \eqn{B_x} and those of \eqn{B_y}. There are \code{R^2} such MC-simulated dissimilarities for each pair of observed sequences.
#'
#' By default, MC standard errors are computed for distances between unique sequences and results are then expanded to all sequences. In addition, results for pairs of identical sequences are expanded to all such pairs in \code{seqdata}. With \code{unique=FALSE}, the computation is redone for each identical pairs and, therefore, results can vary across such identical pairs. Setting \code{unique=TRUE} (default) can save much computation time when same sequences occur multiple times.
#'
#' A progress bar is displayed when \code{verbose=TRUE}. However, the progress bar works only with option \code{snow=TRUE} for parallel computing.
#'
#' \code{seqdistMCSE} is much slower than \code{\link{MCseqdistSE}}, which considers only distances within sets of replicated sequences (generated with \code{\link{MCseqReplicate}}) instead of all combinations of replicated sequences.
#'
#' @param seqdata A state sequence \code{stslist} object as generated by \code{\link[TraMineR]{seqdef}}.
#' @param method Character string. Dissimilarity measure to compute distances. Default is \code{"LCS"}. See \code{\link[TraMineR]{seqdist}}.
#' @param J Integer or vector of positive numbers. If an integer, maximal timing error (number of unit periods around first state of new spell. Default is \code{J=1}.) If a vector, weights of the timing errors \code{k = -K, -(K-1), ..., K-1, K}, where \code{2K + 1} is the length of \code{J}. The vector length must be odd.
#' @param R Integer. Number of random replications of each sequence. Default is \code{R=50}.
#' @param replic Character string. One of \code{"by.pair"} (default) and \code{"once"}. When \code{"by.pair"} a fresh set of replications of the sequences is generated for each distance. When "once", replications of each sequence are generated once and reused for each distance involving the sequence.
#' @param jfixed Logical. Should same error j be applied to all transitions in a sequence? Default is \code{FALSE}.
#' @param kchanges Integer, string, or \code{NULL}. If integer, number of transitions whose time can potentially be altered. If \code{"rand"}, the number of potential changes is randomly selected for each sequence. If \code{NULL} (default), all transitions can potentially be altered.
#' @param verbose Logical. Should waiting and timing messages be printed?
#' @param core Integer. Number of cores to use for parallel computation.
#' @param unique Logical. Should simulations for distances between identical pairs of sequences be run only once? Default is \code{TRUE}.
#' @param model String. Time alteration model. One of \code{"keep.dss"} (default), \code{"indep"} (suppress spells erased by move of transition), and \code{"relative"} (keep time until next transitions unchanged).
#' @param ratios Logical. Should standardized ratios and the standard error of mean simulated distances be returned? Default is \code{TRUE}.
#' @param snow Logical. If \code{TRUE}, \code{doSNOW} is used for parallel computing, otherwise \code{doParallel} is used.
#' @param ... Further arguments passed to \code{\link[TraMineR]{seqdist}}.
#'
#' @return A list of class \code{distMC} with for each pairwise distance: \cr - \code{MC.mean} (\code{dist} object) MC means of distances between MC-replicated sequences,\cr - \code{MC.se} (\code{dist} object) MC standard deviations of distances between MC-replicated sequences, \cr - \code{args.dist} list of arguments  passed to \code{seqdist}, \cr - \code{diss.o} (\code{dist} object) observed distances between sequences, \cr and when \code{ratios = TRUE}: \cr - \code{diss.z} (\code{dist} object) ratios \code{diss.u/MC.se}, \cr - \code{MC.mean.z} (\code{dist} object) ratios  \code{MC.mean/mean.se}, \cr - \code{mean.se} (\code{dist} object) standard errors of \code{MC.mean}.
#'
#' @references
#' Liao, T.F. and G. Ritschard (2023). Evaluating uncertainty of dissimilarity measures between state sequences.
#' Manuscript in preparation.
#'
#'
#' @author Gilbert Ritschard
#'
#' @export
#'
## @exportS3Method print distMC
## @exportS3Method summary distMC
#'
#' @import TraMineR
#' @importFrom WeightedCluster wcAggregateCases
#' @importFrom doParallel registerDoParallel
#' @importFrom parallel makePSOCKcluster stopCluster makeCluster detectCores
#' @importFrom doSNOW registerDoSNOW
#' @importFrom foreach foreach %dopar% %do% %:%
#' @importFrom stats as.dist runif sd
#' @importFrom utils packageVersion txtProgressBar setTxtProgressBar
#' @importFrom iterators icount
#'
#' @seealso \code{\link{MCseqdistSE}}, \code{\link{print.distMC}}, \code{\link{summary.distMC}}, and \code{\link{MCratios}}
#'
#' @examples
#' ## mini test data, 6 sequences of length 4, 4 unique sequences
#' exdata <- read.table(text="
#'                 a a b b
#'                 a a b b
#'                 b b a a
#'                 a c c b
#'                 b b a c
#'                 b b a c
#'                 ")
#' weights=rep(1, nrow(exdata))
#' s.exdata <- seqdef(exdata, weights = weights, id=paste("id",1:nrow(exdata), sep=""))
#'
#' ## Here we call function seqdistMCSE
#' MCd <- seqdistMCSE(s.exdata, method="LCS", J=1, R=50, core=1, verbose=TRUE)
#'
#' ## Results for distances between first sequences
#' MCd
#'
#' ## Summary statistics refer to all distances between original sequences
#' summary(MCd)


seqdistMCSE <- function(seqdata, method="LCS", J=1, R=50, replic="by.pair", verbose=TRUE, core=1, unique=TRUE, model="keep.dss", jfixed=FALSE, kchanges=NULL, ratios=TRUE, snow=TRUE, ...){

  Tver <- packageVersion("TraMineR")
  if (! Tver > "2.2.8")
    cat("!!! You have TraMineR v", paste(Tver), ", some MC.means can be zero because they cannot be computed with v <= 2.2.8!!\n")

  ## requires TraMineR and WeightedCluster

  ## seqdata: an stslist object
  ## method: the metric to use (see seqdist)
  ## core: number of cores for parallel processing
  ## ...: further parameters passed to seqdist

  indel <- list(...)[["indel"]]
  sm <- list(...)[["sm"]]
  #print(length(indel))
  #print(dim(sm))

  if (model=="keep.dss")
    ch.meth <- 1
  else if (model=="indep")
    ch.meth <- 2
  else if (model=="relative")
    ch.meth <- 3
  else
    stop("Unkown model value")

  if (is.character(kchanges) ){
    if (kchanges=="rand")
      kchanges <- -1
    else
      stop("Invalid kchanges value!")
  }

  if(length(J) == 1) {
    JJ <- 2*J + 1
    Jprob <- J
  }
  else {
    JJ <- length(J)
    if (! JJ %% 2)
      stop("Length of J must be odd!")
    if (any(J<0))
      stop("Negative values in J")

    Jprob <- J/sum(J)
    J <- (JJ - 1)/2
  }


  gc(FALSE)
  if (verbose) {
      cat("Computing MC standard errors using ",core," core(s). Please wait ...")
  #   ptime.begin <- proc.time()
  }

  if(!inherits(seqdata,"stslist"))
    stop("argument seqdata must be a stslist object!")


  ## extract unique sequences (we do not need weights here)

  if(unique) {
    aggCases <- wcAggregateCases(seqdata, weights=attr(seqdata,"weights"))
    ## u.seqdata is set of unique sequences
    u.seqdata <- seqdata[aggCases[["aggIndex"]],]
    sdx <- aggCases[["disaggIndex"]]
    names(sdx) <- rownames(seqdata)
  } else {
    aggCases <- NULL
    u.seqdata <- seqdata
    sdx <- 1:nrow(seqdata)
  }
  ## Transform into numeric to avoid issues with unwanted character in state labels
  #DataU <- seqasnum.one(u.seqdata) ## returns numeric for missing
  DataU <- seqnum(u.seqdata) ## returns a stslist
  #print(DataU)
  L <- ncol(DataU)  ## sequence length if all sequences are of equal length
  NU <- nrow(DataU)

  if (NU < 2) {
    uniq <- ifelse (unique, " unique ", " ")
    stop("Only one",uniq,"sequence, should be at least two!")
  }
  #print(dim(DataU))

  sDataU <- DataU ## DataU is stslist
  sdss <- seqdss(sDataU)
  sdur <- seqdur(sDataU)
  if (replic=="once"){
    replset <- MClistrepset(dss=sdss, sdur=sdur, Jprob=Jprob, R=R, ch.meth=ch.meth, jfixed=jfixed, kchanges=kchanges)
  } else {replset <- NULL}

  syst <- system.time(MC.res <- MC.mean.se(sdss, sdur, sDataU,
                                           Jprob=Jprob, R=R, ch.meth=ch.meth,
                                           jfixed = jfixed, kchanges=kchanges,
                                           replset=replset,
                                           core=core, snow=snow, verbose=verbose,
                                           method=method, ...))

  MC.res.o <- MCexpand(MC.res[["MC.mean"]],MC.res[["MC.se"]],sdx=sdx)
  MC.mean <- MC.res.o[["MC.mean"]]
  MC.se <- MC.res.o[["MC.se"]]

  if (verbose) {
    cat(" OK\n")
    message("Computation time:")
    print(syst)
  #
  #   ptime.end <- proc.time()
  #   time.begin <- as.POSIXct(sum(ptime.begin[1:2]), origin = "1960-01-01")
  #   time.end <- as.POSIXct(sum(ptime.end[1:2]), origin = "1960-01-01")
  #   time.elapsed <- format(round(difftime(time.end, time.begin), 3))
  #
  #   if (core>1) message("!! Unreliable elapsed time because of parallel computing !!")
  #   message("elapsed time:", time.elapsed)
  }

  ## MCse for dissimilarities between identical sequences (unique sequences with freq > 1)

  if (unique) {
    um <- which(table(sdx)>1)
    if (!is.null(um) & length(um)>1){
      if (verbose) {
        cat("Computing MC standard errors for dissimilarities between identical sequences ...")
        #   ptime.begin <- proc.time()
      }

      syst <- system.time(MC.res.u <- MC.mean.se.u(sdss, sdur, sDataU, um = um,
                                               Jprob=Jprob, R=R, ch.meth=ch.meth,
                                               jfixed = jfixed, kchanges=kchanges,
                                               replset=replset,
                                               method=method, ...))

      MC.mean <- as.matrix(MC.mean)
      MC.se <- as.matrix(MC.se)
      for (ii in 1:length(um)){
        iii <- which(sdx==um[ii])
        iii
        il <- length(iii)
        for (i in 1:(il-1)){
          for (j in i:il){
            MC.mean[iii[i],iii[j]] <- MC.mean[iii[j],iii[i]] <- MC.res.u[["MC.mean"]][ii]
            MC.se[iii[i],iii[j]] <- MC.se[iii[j],iii[i]] <- MC.res.u[["MC.se"]][ii]
          }
        }
      }
      MC.mean <- as.dist(MC.mean)
      MC.se <- as.dist(MC.se)

      if (verbose) {
        cat(" OK\n")
        #message("Computation time:")
        print(syst)
      }


    }
  } #else MC.res.u <- NULL


  args.dist <- list(...)
  args.dist <- c(list(seqdata = seqdata, full.matrix=FALSE), args.dist)
  args.dist[["method"]] <- method

  suppressMessages(diss.o <- do.call(seqdist, args.dist ))

  # ret <- list(MC.mean = MC.res[["MC.mean"]], MC.se = MC.res[["MC.se"]], R = R,
  #           sdx = sdx, aggCases=aggCases,
  #           u.seq = u.seqdata, args.dist = args.dist,
  #           diss.u = diss.u, MC.res.u = MC.res.u
  #            )

  ret <- list(MC.mean = MC.mean, MC.se = MC.se, R = R,
              #sdx = sdx, aggCases=aggCases,
              weights = attr(seqdata,"weights"),
              args.dist = args.dist,
              diss.o = diss.o
  )

  class(ret) <- c(class(ret),"distMC")

  if (ratios){
    ret <- c(ret, MCratios(ret))
    class(ret) <- c(class(ret),"distMC")
  }

  return(ret)

}


seqMCset <- function(dss,sdur,Jprob,R=10,ch.meth,
                     jfixed, kchanges){
  mdss <- as.matrix(dss)
  ncs <- ncol(mdss)
  set1 <- set2 <- matrix(nrow=R,ncol=ncs)
  #dur1NA <- is.na(sdur[1,])
  #dur2NA <- is.na(sdur[2,])
  for (i in 1:R){
    if (ch.meth==1)
      dur <- t(apply(sdur,1,ch.dur,Jprob=Jprob, jfixed = jfixed, k=kchanges))
    else if (ch.meth==2)
      dur <- t(apply(sdur,1,ch.dur.indep,Jprob=Jprob, jfixed = jfixed, k=kchanges))
    else if (ch.meth==3)
      dur <- t(apply(sdur,1,ch.dur.relat,Jprob=Jprob, jfixed = jfixed, k=kchanges))

    if (nrow(dur)==1) dur <- t(dur)
    wdur1 <- which(dur[1,]>0)
    wdur2 <- which(dur[2,]>0)
    set1[i,wdur1] <- paste0(mdss[1,wdur1],"/",dur[1,wdur1])
    set2[i,wdur2] <- paste0(mdss[2,wdur2],"/",dur[2,wdur2])
    #set1[i,dur1NA] <- NA
    #set2[i,dur2NA] <- NA
    #set1[i,dur[1,]<=0] <- NA
    #set2[i,dur[2,]<=0] <- NA
  }
  #print(set1[1:3,])
  #print(set2[1:3,])
  sts1 <- seqformat(set1, from="SPS", to="STS", SPS.in=list(xfix="", sdsep="/"), stsep="-")
  sts2 <- seqformat(set2, from="SPS", to="STS", SPS.in=list(xfix="", sdsep="/"), stsep="-")
  rownames(sts1) <- paste0(1:R,"B1")
  rownames(sts2) <- paste0(1:R,"B2")

  #print(length(alphabet(dss)))
  seq2sets <- suppressMessages(
    seqdef(rbind(sts1,sts2), alphabet=alphabet(dss))
  )
  #print(length(alphabet(seq2sets)))
  #print("seqdef done!")
  return(seq2sets)
}

seqdistsple <- function(seq2sets, method, what="distMC", ...){

  #indel <- list(...)[["indel"]]
  #sm <- list(...)[["sm"]]
  #print(length(indel))
  #print(dim(sm))
  #print(length(alphabet(seq2sets)))

  MCdi <- suppressMessages(seqdist(seq2sets,
                                    refseq=list(1:(nrow(seq2sets)/2),(nrow(seq2sets)/2+1):nrow(seq2sets)),
                                    method=method,
                                    ...))
  distMC <- list(mean = mean(MCdi), std = sd(MCdi), distMC = MCdi)
  class(distMC) <- c(class(distMC),"distMC")
  if (what=="distMC") return(distMC)
  else return(MCdi)
}


MC.mean.se <- function(dss, sdur, sDataU, k=nrow(dss),
                       jfixed = FALSE, kchanges=NULL,
                       Jprob, R=50, ch.meth=1, replset=NULL,
                       core=1, snow=FALSE, verbose=TRUE,
                       method, ...){
  meth.with.issues <- method %in% c("SVRspell","NMSMST","OMstran","OMspell") & !packageVersion("TraMineR") > "2.2.8"
  if (verbose & (core==1 | snow)) {
    cat("\n")
    progress <- function(n) setTxtProgressBar(pb, n)
    opts <- list(progress = progress)
    pb <- txtProgressBar(min=1, max=k*(k-1)/2, initial=1, style=3)
  }
  else opts <- list()
  kk <- k*(k-1)/2
  ijtab <- getijtable(k)
  if (!is.null(replset))
    idbase <- seq(0,(R-1)*(nrow(dss)), by=nrow(dss))

  if (core==1) {
    MC.mean <- MC.se <- double(kk)
    for (i in 1:(k-1)) {
      for (j in (i+1):k) {
        if (verbose) setTxtProgressBar(pb, didx(i,j,k))
        if (all(seqlength(dss[c(i,j),,drop=FALSE])==1)){ ## SE is zero
          if (!meth.with.issues){
            ##cat("dss of length 1, (i,j)= (",i,",",j,")\n")
            MC.mean[didx(i,j,k)] <- suppressMessages(seqdist(sDataU[c(i,j),],method=method,...)[1,2])
            ##cat("MC.mean = ", MC.mean[didx(i,j,k)])
          } else {
            cat("\n !!! Special case (",i,",",j,") requires TraMineR > 2.2.8, MC.mean set as zero \n")
            MC.mean[didx(i,j,k)] <- 0
          }
          MC.se[didx(i,j,k)] <- 0
        } else {
          if (!is.null(replset)){
            seq2sets <- replset[c(idbase+i,idbase+j),]
          } else {
            seq2sets <- seqMCset(dss[c(i,j),,drop=FALSE],sdur[c(i,j),,drop=FALSE],
                                 Jprob=Jprob, R=R, ch.meth=ch.meth,
                                 jfixed = jfixed, kchanges=kchanges)

          }
          distMC <- seqdistsple(seq2sets, method=method, what="distMC", ...)
          MC.mean[didx(i,j,k)] <- distMC[["mean"]]
          MC.se[didx(i,j,k)] <- distMC[["std"]]
        }
      }
    }
    if (verbose) close(pb)
  } else { ## parallel computing
    #pb <- txtProgressBar(min=1, max=k-1, initial=1, style=3)
    #cl <- makePSOCKcluster(core)
    cl <- makeCluster(core, type="SOCK")
    if (snow)
      registerDoSNOW(cl)
    else
      registerDoParallel(cl)

    #mat.d <- foreach (i = icount(k-1), .combine=rbind) %:%
    #  foreach (j = icount(k-1), .combine=rbind, .packages=c("TraMineR"), .options.snow=opts) %dopar% {
    #  if (j>k-i)
    #    return(NULL)
    #  else {...
    #  }

    pij <- integer(1)

    mat.d <- foreach (pij = icount(kk), .combine=rbind, .packages=c("TraMineR"), .options.snow=opts) %dopar% {
      i <- ijtab[pij,1]
      jj <- ijtab[pij,2]
      #jj <- j + i
      if (all(seqlength(dss[c(i,jj),,drop=FALSE])==1)){ ## SE is zero
        if (!meth.with.issues){
          mean <- suppressMessages(seqdist(sDataU[c(i,jj),],method=method,...)[1,2])
        } else {
          mean <- 0
        }
        distMC <- list(mean = mean, std = 0)
      } else {
        if (!is.null(replset)){
          seq2sets <- replset[c(idbase+i,idbase+jj),]
        } else {
          seq2sets <- seqMCset(dss[c(i,jj),,drop=FALSE],sdur[c(i,jj),,drop=FALSE],
                               Jprob=Jprob, R=R, ch.meth=ch.meth,
                               jfixed = jfixed, kchanges=kchanges)

        }
        distMC <- seqdistsple(seq2sets, method=method, ...)
      }
      return(c(distMC[["mean"]],distMC[["std"]]))
    }

    stopCluster(cl)
    if(verbose & snow) {
      close(pb)
    }

    mat.d <- as.matrix(mat.d)
    if(ncol(mat.d)<2) mat.d <- t(mat.d)
    #print(dim(mat.d))
    MC.mean <- mat.d[,1,drop=FALSE]
    MC.se <- mat.d[,2,drop=FALSE]
  }

  ## Making MC.mean a dist object
  class(MC.mean) <- "dist"
  attr(MC.mean, "Size") <- k
  attr(MC.mean, "Labels") <- rownames(dss)
  attr(MC.mean, "Diag") <- FALSE
  attr(MC.mean, "Upper") <- FALSE

  ## Making MC.se a dist object
  class(MC.se) <- "dist"
  attr(MC.se, "Size") <- k
  attr(MC.se, "Labels") <- rownames(dss)
  attr(MC.se, "Diag") <- FALSE
  attr(MC.se, "Upper") <- FALSE

  return(list(MC.mean=MC.mean, MC.se=MC.se, R=R, J=Jprob))
}

## for dissimilarities between identical sequences
MC.mean.se.u <- function(dss, sdur, sDataU, um,
                         Jprob, R=50, ch.meth, replset=NULL,
                         jfixed = FALSE, kchanges=NULL,
                         method, ...){
  #meth.with.issues <- method %in% c("SVRspell","NMSMST","OMstran","OMspell") & !packageVersion("TraMineR") > "2.2.8"

  k <- length(as.vector(um))
  if (!is.null(replset))
    idbase <- seq(0,(R-1)*(nrow(dss)), by=nrow(dss))

  # if (verbose & (core==1 | snow)) {
  #   cat("\n")
  #   progress <- function(n) setTxtProgressBar(pbb, n)
  #   opts <- list(progress = progress)
  #   pbb <- txtProgressBar(min=1, max=k, initial=1, style=3)
  # }
  # else
  #opts <- list()

  #if (core==1) {
    MC.mean <- MC.se <- double(k)
    for (i in 1:k) {
      #if (verbose) setTxtProgressBar(pbb, i)
      if (all(seqlength(dss[um[i],,drop=FALSE])==1)){ ## Mean and SE is zero
        MC.mean[i] <- 0
        MC.se[i] <- 0
      } else {
        if (!is.null(replset)){
          seq2sets <- replset[c(idbase+um[i],idbase+um[i]),,drop=FALSE]
        } else {
          seq2sets <- seqMCset(dss[c(um[i],um[i]),,drop=FALSE],sdur[c(um[i],um[i]),,drop=FALSE],
                             Jprob=Jprob, R=R, ch.meth=ch.meth,
                             jfixed = jfixed, kchanges=kchanges)
        }
        distMC <- seqdistsple(seq2sets, method=method, what="distMC", ...)
        MC.mean[i] <- distMC[["mean"]]
        MC.se[i] <- distMC[["std"]]
      }
    }
    #if (verbose) close(pbb)
  # } else { ## parallel computing
  #   #cl <- makePSOCKcluster(core)
  #   cl <- makeCluster(core, type="SOCK")
  #   if (snow)
  #     registerDoSNOW(cl)
  #   else
  #     registerDoParallel(cl)
  #
  #   mat.d <- foreach (i = icount(k), .combine=rbind, .packages=c("TraMineR"), .options.snow=opts) %dopar% {
  #
  #       if (seqlength(dss[um[i],,drop=FALSE])==1){ ## mean and SE is zero
  #         distMC <- list(mean = 0, std = 0)
  #       } else {
  #         seq2sets <- seqMCset(dss[c(um[i],um[i]),,drop=FALSE],sdur[c(um[i],um[i]),,drop=FALSE], Jprob=Jprob, R=R)
  #         distMC <- seqdistsple(seq2sets, method=method, ...)
  #       }
  #       c(distMC[["mean"]],distMC[["std"]])
  #     }
  #   stopCluster(cl)
  #   if(verbose & snow) {
  #     close(pbb)
  #   }
  #
  #     mat.d <- as.matrix(mat.d)
  #     if(ncol(mat.d)<2) mat.d <- t(mat.d)
  #     #print(dim(mat.d))
  #     MC.mean <- mat.d[,1,drop=FALSE]
  #     MC.se <- mat.d[,2,drop=FALSE]
  #}

  return(list(MC.mean=MC.mean, MC.se=MC.se, R=R, J=Jprob))
}



MCexpand <- function(MC.mean, MC.se, sdx){
  MCdiss.ori <- as.matrix(MC.mean)[sdx,sdx]
  sd.ori <- as.matrix(MC.se)[sdx,sdx]
  dimnames(MCdiss.ori) <-  dimnames(sd.ori) <- list(names(sdx),names(sdx))
  res <- list(MC.mean = as.dist(MCdiss.ori), MC.se = as.dist(sd.ori))
  return(res)
}



MCdistrib <- function(dss,dur,id2=1:2, Jprob, R=100, model="keep.dss",
                      jfixed=FALSE, kchanges=NULL,
                      method="LCS", ...){
  Jprob <- Jprob/sum(Jprob)
  if (model=="keep.dss")
    ch.meth <- 1
  else if (model=="indep")
    ch.meth <- 2
  else if (model=="relative")
    ch.meth <- 3
  else
    stop("Unkown model value")

  #seq2sets <- seqdistMCSE:::seqMCset(dss[id2,],dur[id2,], Jprob=Jprob, R=R)
  #MC.dist <- seqdistMCSE:::seqdistsple(seq2sets, method=method, what="MC distances", ...)
  seq2sets <- seqMCset(dss[id2,],dur[id2,], Jprob=Jprob, R=R, ch.meth=ch.meth,
                       jfixed=jfixed, kchanges=kchanges)
  MC.dist <- seqdistsple(seq2sets, method=method, what="MC distances", ...)
  return(MC.dist)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCratios.R
\name{MCratios}
\alias{MCratios}
\title{Ratios of distances on their standard errors}
\usage{
MCratios(object, diss.o = NULL)
}
\arguments{
\item{object}{Object of class \code{distMC} as generated by \code{\link{MCseqdistSE}} and \code{\link{seqdistMCSE}}.}

\item{diss.o}{Matrix or \code{dist} object. Pairwise dissimilarities between observed sequences.}
}
\value{
\code{diss.z}, \code{MC.mean.z}, and \code{mean.se} (the three as \code{dist} objects).
}
\description{
Ratios of the observed distances to their MC standard errors and of the mean MC-simulated distances to the standard error of the mean.
}
\details{
The standard error of the mean simulated distances is \code{mean.se = MC.se/sqrt(R)} (or \code{mean.se = MC.sd/R} when object is obtained with \code{seqdistMCSE}, because there are \code{R*R} simulated distances in that case). The ratios computed are \code{diss.z = diss.o/MC.se}, where \code{diss.o} is the distance between observed sequences, and \code{MC.mean.z = MC.mean/mean.se} with \code{MC.mean} the mean of the MC-simulated distances.

When \code{diss.o=NULL}, the \code{diss.o} element of \code{object} is used when it exists.

This function is handy to get afterwards ratios for outcome of \code{\link{seqdistMCSE}} obtained with \code{ratios=FALSE}.
}
\seealso{
\code{\link{MCseqdistSE}} and \code{\link{print.MCratios}}.
}
\author{
Gilbert Ritschard
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCpj.R
\name{MCpj}
\alias{MCpj}
\title{Generate distribution of timing errors}
\usage{
MCpj(Emean, pzero = NULL, maxterr = 10, pinterv = 0.99, fill.short.side = TRUE)
}
\arguments{
\item{Emean}{scalar or vector of size two. Expected size of non-zero timing errors. If a vector, the first value is used for negative errors and the second value for positive errors. If a scalar, the value is used for both negative and positive errors. Values must be strictly greater than 1.}

\item{pzero}{number in range [0,1]. Probability of no-error. If \code{NULL} (default), \code{pzero} is set to the the greatest probability of zero between the right and left side Poisson distributions.}

\item{maxterr}{integer. Maximal error size to consider. Default is 10.}

\item{pinterv}{control value used for solving numerically an implicit function. Default is .99 and should be increased in case the zero of the implicit function cannot be found because of ending values of same sign.}

\item{fill.short.side}{logical. Should the shortest side be filled with zeros to equal length of the other side. Default is \code{TRUE}.}
}
\value{
The vector of probabilities Pj with the computed \code{lambda} values as attribute.
}
\description{
Generates a distribution of timing errors that complies with the provided expected size of non-zero timing errors and the expected probability of no error.
}
\details{
Currently \code{\link{MCseqReplicate}} expects a vector Pj with same number of backward and forward error values. To comply with this, the shorter side of Pj is by default filled with zeros.
}
\examples{
# expected timing error of 1.2 on each side
MCpj(Emean=1.2, pzero=.4)

# expected backward timing error higher than for forward errors
MCpj(Emean=c(3.5,1.2), pzero=.4)


}
\seealso{
\code{\link{MCseqReplicate}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCmethods.R
\name{summary.distMC}
\alias{summary.distMC}
\title{Summary method for distMC objects}
\usage{
\method{summary}{distMC}(object, ..., silent = FALSE)
}
\arguments{
\item{object}{\code{distMC} object as returned by \code{MCseqdistSE}.}

\item{...}{further arguments passed to or from other methods.}

\item{silent}{logical: Should additional info be displayed?}
}
\value{
\code{fivenumb} table with the statistics (min, Q1, med, Q3, max) of the observed dissimilarities, the mean, standard deviation, and standard error of the MC-simulated dissimilarities, standardized ratios, MC-bias and mean squared errors when available.
}
\description{
Prints summary statistics of the observed dissimilarity \code{diss}, the mean \code{MC.mean}, standard deviation \code{MC.sd}, and standard error of dissimilarities between MC-replicated sequences, and the ratios  \code{diss/MC.se} and \code{MC.mean/MC.se}. Reported statistics concern all distances between original sequences.
}
\seealso{
\code{\link{MCseqdistSE}}, \code{\link{print.distMC}}
}
\author{
Gilbert Ritschard
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCmethods.R
\name{print.distMC}
\alias{print.distMC}
\title{Print method for distMC objects}
\usage{
\method{print}{distMC}(x, n = 6, what = "all", ...)
}
\arguments{
\item{x}{\code{distMC} object as returned by \code{MCseqdistSE}.}

\item{n}{Integer. Number of first sequences. Default is 6. If \code{n==0} or there are less than \code{n} sequences, results are printed for all pairs of sequences.}

\item{what}{character string. One of \code{"mean"}, \code{"sd"}, \code{"bias"}, \code{"both"}, and \code{"all"} (default). When \code{"all"}, ratios, when present are printed together with the mean and standard deviation. When \code{"both"}, means and standard deviations are printed.}

\item{...}{further arguments passed to or from other methods.}
}
\value{
Last printed table, a \code{matrix} when \code{toref} attribute is \code{TRUE} and a \code{dist} object otherwise.
}
\description{
Prints, for each pair of the first \code{n} sequences, the mean and/or the standard deviation of the MC-replicated distances between sequences. When available, ratios are also printed by default.
}
\seealso{
\code{\link{MCseqdistSE}}, \code{\link{summary.distMC}}.
}
\author{
Gilbert Ritschard
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCudist.R
\name{MCudist}
\alias{MCudist}
\title{Dissimilarities between unique replicated sequences}
\usage{
MCudist(MCrseqdata, method = "LCS", seqref = NULL, ...)
}
\arguments{
\item{MCrseqdata}{list of replicated \code{stslist} state sequence datasets (all of same size and with same alphabet)}

\item{method}{string. Name of distance method (see \code{seqdist}).}

\item{seqref}{state sequence object of class \code{stslist}. Fixed reference sequences.}

\item{...}{Further arguments passed to \code{seqdist}}
}
\value{
object of class \code{u.diss} (pairwise dissimilarities between unique sequences) with two attributes: \code{sdx}, inverted aggregation indexes, \code{N}, number of datasets, and \code{obs}, logical indicating whether \code{k=N} corresponds to observed sequences.
}
\description{
Returns the dissimilarity matrix (or \code{dist} object) between merged replicated sequences with the disaggregation indexes as attribute.
}
\examples{
## mini test data, 6 sequences of length 4, 4 unique sequences
exdata <- read.table(text="
                a a b b
                a a b b
                b b a a
                a c c b
                b b a c
                b b a c
                ")
weights=rep(1, nrow(exdata))
s.exdata <- seqdef(exdata, weights = weights, id=paste("id",1:nrow(exdata), sep=""))

## 3 altered sequence datasets
(altseq.list <- MCseqReplicate(s.exdata, J=1, R=3))

MCnunique(altseq.list, check=TRUE)


u.diss <- MCudist(altseq.list, method="LCS", full.matrix=FALSE)
## Dissimilarities within first MC-set
MCExtractDist(u.diss, 1)

## list of dissimilarity matrices
disslist <- MCdisslist(altseq.list, use.udiss=TRUE)

}
\seealso{
\code{\link{MCExtractDist}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCmethods.R
\name{summary.MCratios}
\alias{summary.MCratios}
\title{Summary method for MCratios objects}
\usage{
\method{summary}{MCratios}(object, ..., weights = NULL, silent = FALSE, thresh = 2)
}
\arguments{
\item{object}{\code{MCratios} object as returned by \code{MCratios}.}

\item{...}{further arguments passed to or from other methods.}

\item{weights}{vector of doubles. Case weights.}

\item{silent}{logical: Should additional info be displayed?}

\item{thresh}{real: threshold for counting ratios less than \code{thresh}}
}
\value{
\code{fivenumb} table with the statistics (min, Q1, med, Q3, max) of \code{mean.se} and the standardized ratios \code{diss.z} and \code{MC.mean.z}.
}
\description{
Prints summary statistics of the ratios  \code{diss/MC.se} and \code{MC.mean/MC.se}. Reported statistics concern all distances between original sequences.
}
\seealso{
\code{\link{MCseqdistSE}}, \code{\link{print.distMC}}
}
\author{
Gilbert Ritschard
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCseqReplicate.R
\name{MCseqReplicate}
\alias{MCseqReplicate}
\title{Generate R altered sequence data sets.}
\usage{
MCseqReplicate(
  seqdata,
  J = 1,
  R = 20,
  silent = FALSE,
  unique = FALSE,
  model = "keep.dss",
  jfixed = FALSE,
  kchanges = NULL,
  include.obs = FALSE
)
}
\arguments{
\item{seqdata}{A state sequence \code{stslist} object as generated by \code{\link[TraMineR]{seqdef}}.}

\item{J}{Integer or vector of positive numbers. If an integer, maximal timing error (number of unit periods around first state of new spell. Default is \code{J=1}. If a vector, weights of the timing errors \code{k = -K, -(K-1), ..., K-1, K}, where \code{2K + 1} is the length of \code{J}. The vector length must be odd.}

\item{R}{Integer. Number of random replicated sequence data. Default is \code{R=10}. The weights are internally normalized to sum to 1.}

\item{silent}{Logical. Should waiting and timing messages be hidden?}

\item{unique}{Logical. Should only unique sequences be replicated? Default is \code{FALSE}. If \code{TRUE} weights will reflect the multiple occurrences of each original unique sequence.}

\item{model}{String. Time alteration model. One of \code{"keep.dss"} (default), \code{"indep"} (suppress spells erased by move of transition), and \code{"relative"} (keep time until next transitions unchanged).}

\item{jfixed}{Logical. Should same error j be applied to all transitions in a sequence? Default is \code{FALSE}.}

\item{kchanges}{Integer, string, or \code{NULL}. If integer, number of transitions whose time can potentially be altered in each sequence. If \code{"rand"}, the number of potential changes is randomly selected for each sequence. If \code{NULL} (default), all transitions can potentially be altered.}

\item{include.obs}{logical. Should the observed sequence data be added as last element.}
}
\value{
List of R altered \code{stslist} objects plus observed sequence object as last element when \code{include.obs=TRUE}.
}
\description{
R \code{stslist} state sequence objects are generated by applying the chosen timing error model to the provided state sequence object.
}
\details{
This function is handy for testing how outcome of a sequence analysis may vary with timing errors in the reported sequences.

Use the vector form of \code{J} to specify the probability distribution of the timing error. See function \code{\link{MCpj}} to generate a probability vector that complies with expected mean timing errors.
}
\examples{
## mini test data, 6 sequences of length 4, 4 unique sequences
exdata <- read.table(text="
                a a b b
                a a b b
                b b a a
                a c c b
                b b a c
                b b a c
                ")
weights=rep(1, nrow(exdata))
s.exdata <- seqdef(exdata, weights = weights, id=paste("id",1:nrow(exdata), sep=""))

## 3 altered sequence datasets
(altseq.list <- MCseqReplicate(s.exdata, J=1, R=3))

## list of dissimilarity matrices
suppressMessages(dist.list <- lapply(altseq.list, seqdist, method="LCS", full.matrix=FALSE))
dist.list

## Can also be obtained with MCdisslist, which offers option use.udiss;
## use.udiss=TRUE is faster when number of unique merged replicated
## sequences is less than n*sqrt(R).
suppressMessages(dist.list <- MCdisslist(altseq.list, method="LCS", use.udiss=TRUE))

## Replication based on expected left and right non-zero errors of 1.1
##  and assuming a 0.5 probability of no error
Pj <- MCpj(Emean=1.1, pzero=.5)
(altseq2.list <- MCseqReplicate(s.exdata, J=Pj, R=3))
}
\references{
Ritschard, G. and Liao, T.F. (2026). Assessing the Impact of Timing Errors in Sequence Analysis. \emph{International Journal of Social Research Methodology}. Forthcoming
}
\seealso{
\code{\link{MCpj}}
}
\author{
Gilbert Ritschard
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seqdistMCSE.R
\name{seqdistMCSE}
\alias{seqdistMCSE}
\title{Mean and standard deviation of dissimilarities between pairs of randomly altered sequences.}
\usage{
seqdistMCSE(
  seqdata,
  method = "LCS",
  J = 1,
  R = 50,
  replic = "by.pair",
  verbose = TRUE,
  core = 1,
  unique = TRUE,
  model = "keep.dss",
  jfixed = FALSE,
  kchanges = NULL,
  ratios = TRUE,
  snow = TRUE,
  ...
)
}
\arguments{
\item{seqdata}{A state sequence \code{stslist} object as generated by \code{\link[TraMineR]{seqdef}}.}

\item{method}{Character string. Dissimilarity measure to compute distances. Default is \code{"LCS"}. See \code{\link[TraMineR]{seqdist}}.}

\item{J}{Integer or vector of positive numbers. If an integer, maximal timing error (number of unit periods around first state of new spell. Default is \code{J=1}.) If a vector, weights of the timing errors \code{k = -K, -(K-1), ..., K-1, K}, where \code{2K + 1} is the length of \code{J}. The vector length must be odd.}

\item{R}{Integer. Number of random replications of each sequence. Default is \code{R=50}.}

\item{replic}{Character string. One of \code{"by.pair"} (default) and \code{"once"}. When \code{"by.pair"} a fresh set of replications of the sequences is generated for each distance. When "once", replications of each sequence are generated once and reused for each distance involving the sequence.}

\item{verbose}{Logical. Should waiting and timing messages be printed?}

\item{core}{Integer. Number of cores to use for parallel computation.}

\item{unique}{Logical. Should simulations for distances between identical pairs of sequences be run only once? Default is \code{TRUE}.}

\item{model}{String. Time alteration model. One of \code{"keep.dss"} (default), \code{"indep"} (suppress spells erased by move of transition), and \code{"relative"} (keep time until next transitions unchanged).}

\item{jfixed}{Logical. Should same error j be applied to all transitions in a sequence? Default is \code{FALSE}.}

\item{kchanges}{Integer, string, or \code{NULL}. If integer, number of transitions whose time can potentially be altered. If \code{"rand"}, the number of potential changes is randomly selected for each sequence. If \code{NULL} (default), all transitions can potentially be altered.}

\item{ratios}{Logical. Should standardized ratios and the standard error of mean simulated distances be returned? Default is \code{TRUE}.}

\item{snow}{Logical. If \code{TRUE}, \code{doSNOW} is used for parallel computing, otherwise \code{doParallel} is used.}

\item{...}{Further arguments passed to \code{\link[TraMineR]{seqdist}}.}
}
\value{
A list of class \code{distMC} with for each pairwise distance: \cr - \code{MC.mean} (\code{dist} object) MC means of distances between MC-replicated sequences,\cr - \code{MC.se} (\code{dist} object) MC standard deviations of distances between MC-replicated sequences, \cr - \code{args.dist} list of arguments  passed to \code{seqdist}, \cr - \code{diss.o} (\code{dist} object) observed distances between sequences, \cr and when \code{ratios = TRUE}: \cr - \code{diss.z} (\code{dist} object) ratios \code{diss.u/MC.se}, \cr - \code{MC.mean.z} (\code{dist} object) ratios  \code{MC.mean/mean.se}, \cr - \code{mean.se} (\code{dist} object) standard errors of \code{MC.mean}.
}
\description{
For each pair of sequences, returns the mean and standard deviation (MCSE) of the dissimilarities between all combinations of MC-replicated sequences, where sequences are replicated with random timing changes.
}
\details{
Let \eqn{B_x} be the set of \code{R} sequences derived from a sequence \eqn{x} by randomly altering the timing of the transitions (state changes) in \eqn{x}. The MC standard error of the dissimilarity \eqn{d(x,y)} between two sequences \eqn{x} and \eqn{y} is the empirical standard deviation of the dissimilarities between the sequences of \eqn{B_x} and those of \eqn{B_y}. There are \code{R^2} such MC-simulated dissimilarities for each pair of observed sequences.

By default, MC standard errors are computed for distances between unique sequences and results are then expanded to all sequences. In addition, results for pairs of identical sequences are expanded to all such pairs in \code{seqdata}. With \code{unique=FALSE}, the computation is redone for each identical pairs and, therefore, results can vary across such identical pairs. Setting \code{unique=TRUE} (default) can save much computation time when same sequences occur multiple times.

A progress bar is displayed when \code{verbose=TRUE}. However, the progress bar works only with option \code{snow=TRUE} for parallel computing.

\code{seqdistMCSE} is much slower than \code{\link{MCseqdistSE}}, which considers only distances within sets of replicated sequences (generated with \code{\link{MCseqReplicate}}) instead of all combinations of replicated sequences.
}
\examples{
## mini test data, 6 sequences of length 4, 4 unique sequences
exdata <- read.table(text="
                a a b b
                a a b b
                b b a a
                a c c b
                b b a c
                b b a c
                ")
weights=rep(1, nrow(exdata))
s.exdata <- seqdef(exdata, weights = weights, id=paste("id",1:nrow(exdata), sep=""))

## Here we call function seqdistMCSE
MCd <- seqdistMCSE(s.exdata, method="LCS", J=1, R=50, core=1, verbose=TRUE)

## Results for distances between first sequences
MCd

## Summary statistics refer to all distances between original sequences
summary(MCd)
}
\references{
Liao, T.F. and G. Ritschard (2023). Evaluating uncertainty of dissimilarity measures between state sequences.
Manuscript in preparation.
}
\seealso{
\code{\link{MCseqdistSE}}, \code{\link{print.distMC}}, \code{\link{summary.distMC}}, and \code{\link{MCratios}}
}
\author{
Gilbert Ritschard
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCudist.R
\name{MCExtractDist}
\alias{MCExtractDist}
\title{Extract k-th dissimilarity matrix from u.diss}
\usage{
MCExtractDist(u.diss, k, full.matrix = FALSE)
}
\arguments{
\item{u.diss}{\code{u.diss} object returned by \code{MCudist}: dissimilarities between unique replicated sequences.}

\item{k}{integer. Subset index number for which the dissimilarity matrix must be extracted}

\item{full.matrix}{logical. If \code{FALSE}, the distance matrix is returned as a \code{dist} object. Ignored for distances to reference sequences.}
}
\value{
a dissimilarity matrix or distance object.
}
\description{
Extract k-th dissimilarity matrix from u.diss
}
\seealso{
\code{\link{MCudist}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCclustqual.R
\name{MCclustqual}
\alias{MCclustqual}
\title{Cluster quality measures by MC-sets}
\usage{
MCclustqual(
  disslist,
  ncluster = 10,
  clustmeth = "PAM",
  weights = NULL,
  core = 1,
  snow = TRUE,
  silent = FALSE,
  ...
)
}
\arguments{
\item{disslist}{List of MC-dissimilarity matrices (or \code{dist} objects).}

\item{ncluster}{integer vector. Range of number of groups. Default is \code{2:10}.}

\item{clustmeth}{character. Clustering method. Either \code{"PAM"} (default) or a \code{stats::\link{hclust}} method.}

\item{weights}{vector of doubles. Case weights. If \code{NULL} (default), equal weights are used.}

\item{core}{Integer. Number of cores for parallel computing.}

\item{snow}{Logical. If \code{TRUE}, \code{doSNOW} is used for parallel computing, otherwise \code{doParallel} is used.}

\item{silent}{Logical. Should waiting and timing messages be hidden?}

\item{...}{Further arguments passed to clustering functions.}
}
\value{
A list with two lists: \code{qual.tab}, list of tables of cluster quality statistics per MC-dissimilarity matrix, and \code{qual.max} list of cluster number $k$ for which the statistics reach their maximum (minimum for HC), \code{max.freq}, the frequency table of maximum over the MC-replicated sets, and \code{qual.obs}, cluster quality indexes for the observed sequences.
}
\description{
Cluster quality measures for a range of number of groups by MC-replicated set.
}
\details{
When \code{attr(MCdisslist,"obs")} is \code{TRUE}, the last element of \code{disslist} is treated as the dissimilarity matrix of the observed sequences.
}
\examples{
## mini test data, 6 sequences of length 4, 4 unique sequences
exdata <- read.table(text="
                a a b b
                a a b b
                b b a a
                a c c b
                b b a c
                b b a c
                ")
weights=rep(1, nrow(exdata))
s.exdata <- seqdef(exdata, weights = weights, id=paste("id",1:nrow(exdata), sep=""))

## 3 altered sequence datasets
set.seed(25)
altseq.list <- MCseqReplicate(s.exdata, J=1, R=3)
## list of dissimilarity matrices
disslist <- MCdisslist(altseq.list, method="LCS")
diss.o <- seqdist(s.exdata, method="LCS")
## cluster per MC-dissimilarity matrices
res <- MCclustqual(disslist,ncluster=3)
res


}
\seealso{
\code{\link[WeightedCluster]{as.clustrange}},  \code{\link[WeightedCluster]{wcKMedRange}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCdisscorr.R
\name{MCdisscorr}
\alias{MCdisscorr}
\title{Correlation between observed and MC-simulated distances}
\usage{
MCdisscorr(disslist, diss.o = NULL, method = "Spearman", weights = NULL)
}
\arguments{
\item{disslist}{List of matrices or dist objects: the MC-replicated dissimilarities}

\item{diss.o}{Matrix or dist object: Observed dissimilarities}

\item{method}{String. One of \code{"Spearman"} (default) and \code{"Pearson"}.}

\item{weights}{vector of doubles. Case weights. If \code{NULL} (default), equal weights are used.}
}
\value{
vector of correlation between observed and MC-dissimilarities.
}
\description{
Correlation between observed and MC-simulated distances
}
\details{
When \code{diss.o=NULL}, the last element of \code{disslist} is taken as \code{diss.o} and the other elements as sets of MC-replicated dissimilarities.
}
\examples{
## mini test data, 6 sequences of length 4, 4 unique sequences
exdata <- read.table(text="
                a a b b
                a a b b
                b b a a
                a c c b
                b b a c
                b b a c
                ")
weights=rep(1, nrow(exdata))
s.exdata <- seqdef(exdata, weights = weights, id=paste("id",1:nrow(exdata), sep=""))

## 3 altered sequence datasets
set.seed(25)
altseq.list <- MCseqReplicate(s.exdata, J=1, R=3, include.obs=TRUE)
## list of dissimilarity matrices
disslist <- MCdisslist(altseq.list)
MCdisscorr(disslist)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCclustcomp.R
\name{MCclustcomp}
\alias{MCclustcomp}
\title{Comparing MC-clusters with cluster of observed data}
\usage{
MCclustcomp(clustlist, clust.o = NULL, weights = NULL)
}
\arguments{
\item{clustlist}{List of MC-replicated vectors of cluster memberships.}

\item{clust.o}{Cluster memberships based on observed dissimilarities.}

\item{weights}{vector of doubles. Case weights. If \code{NULL} (default), equal weights are used.}
}
\value{
A table with in columns the list of comparison scores provided by \code{aricode::\link[aricode]{clustComp}} for each replicated set, except Chi2, which is replaced by Cramer's V.
}
\description{
Comparison indexes between clusters based on observed data and each of MC-replicated clusters.
}
\details{
When \code{diss.o=NULL}, the last element of \code{disslist} is taken as \code{diss.o} and the other elements as sets of MC-replicated dissimilarities.
}
\examples{
## mini test data, 6 sequences of length 4, 4 unique sequences
exdata <- read.table(text="
                a a b b
                a a b b
                b b a a
                a c c b
                b b a c
                b b a c
                ")
weights=rep(1, nrow(exdata))
s.exdata <- seqdef(exdata, weights = weights, id=paste("id",1:nrow(exdata), sep=""))

## 3 altered sequence datasets
set.seed(25)
altseq.list <- MCseqReplicate(s.exdata, J=1, R=3)
## list of dissimilarity matrices
disslist <- MCdisslist(altseq.list, method="LCS")
diss.o <- seqdist(s.exdata, method="LCS")
## cluster per MC-dissimilarity matrices
library(WeightedCluster)
clust.o <- wcKMedoids(diss.o, k=2, cluster.only=TRUE)
clustlist <- lapply(disslist, wcKMedoids, k=2, cluster.only=TRUE)
res <- MCclustcomp(clustlist, clust.o=clust.o)
res


}
\seealso{
\code{\link[aricode]{clustComp}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCmdscorr.R
\name{MCmdscorr}
\alias{MCmdscorr}
\title{Correlation between 1st MDS factor of observed and MC-simulated distances}
\usage{
MCmdscorr(
  disslist,
  diss.o = NULL,
  method = "Spearman",
  weights = NULL,
  what = "corr",
  core = 1,
  snow = TRUE,
  silent = FALSE
)
}
\arguments{
\item{disslist}{List of matrices or dist objects: the MC-replicated dissimilarities}

\item{diss.o}{Matrix or dist object: Observed dissimilarities}

\item{method}{String. One of \code{"Spearman"} (default) and \code{"Pearson"}.}

\item{weights}{vector of doubles. Case weights. If \code{NULL} (default), equal weights are used.}

\item{what}{String. One of \code{"corr"} (correlations, default), \code{"mds"} (list of mds scores), and \code{"both"}.}

\item{core}{Integer. Number of cores for parallel computing.}

\item{snow}{Logical. If \code{TRUE}, \code{doSNOW} is used for parallel computing, otherwise \code{doParallel} is used.}

\item{silent}{Logical. Should waiting and timing messages be hidden?}
}
\value{
when \code{what="corr"}, vector of correlation between mds of dissimilarities in MC-replicated sets, when \code{what="mds"}, of first mds scores, and when \code{what="both"}, list with \code{corr} as first element and \code{mdslist}, the list of mds scores as second element.
}
\description{
Correlation between 1st MDS factor of observed and MC-simulated distances
}
\details{
When \code{diss.o=NULL}, the last element of \code{disslist} is taken as \code{diss.o} and the other elements as sets of MC-replicated dissimilarities.
}
\examples{
## mini test data, 6 sequences of length 4, 4 unique sequences
exdata <- read.table(text="
                a a b b
                a a b b
                b b a a
                a c c b
                b b a c
                b b a c
                ")
weights=rep(1, nrow(exdata))
s.exdata <- seqdef(exdata, weights = weights, id=paste("id",1:nrow(exdata), sep=""))

## 3 altered sequence datasets
set.seed(25)
altseq.list <- MCseqReplicate(s.exdata, J=1, R=3)
## list of dissimilarity matrices
disslist <- MCdisslist(altseq.list)
MCmdscorr(disslist)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCdisslist.R
\name{MCdisslist}
\alias{MCdisslist}
\title{List of dissimilarity matrices}
\usage{
MCdisslist(
  MCrseqdata,
  method = "LCS",
  seqref = NULL,
  full.matrix = FALSE,
  use.udiss = FALSE,
  ...
)
}
\arguments{
\item{MCrseqdata}{List of state sequence objects of class \code{stslist}.}

\item{method}{string. Name of a distance method (see \code{\link[TraMineR]{seqdist}}).}

\item{seqref}{state sequence object of class \code{stslist}. Fixed reference sequences.}

\item{full.matrix}{logical. Should pairwise distances be returned in matrix form? If \code{FALSE} (default),
a list of \code{dist} objects is returned. Applies only when \code{seqref=NULL}.}

\item{use.udiss}{logical. Should computation be based on unique sequences?}

\item{...}{further arguments passed to \code{seqdist}.}
}
\value{
list of dissimilarity matrices or \code{dist} objects with logical attribute \code{"obs"}, which is \code{TRUE} when the list includes the dissimilarities between observed sequences as last element.
}
\description{
Compute the dissimilarity matrix for each of the provided sets of sequences.
}
\details{
When \code{use.udiss=TRUE}, the function first computes dissimilarities between unique merged replicated sequences through a single call to \code{seqdist()} and the set of dissimilarity matrices are then extracted from the resulting distance matrix. This is generally faster when the number of unique merged replicated sequences is less than \code{sqrt(number of replicated datasets) * (sample size)}, which can be checked with \code{\link{MCnunique}}.
}
\seealso{
\code{\link{MCseqReplicate}}, \code{\link{MCudist}} and examples in their help pages.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCseqdistSE.R
\name{MCseqdistSE}
\alias{MCseqdistSE}
\title{Distance standard errors derived from sets of MC-replicated sequences}
\usage{
MCseqdistSE(
  dissrepl = "LCS",
  MCrseqdata = NULL,
  udiss = FALSE,
  full.matrix = FALSE,
  ...
)
}
\arguments{
\item{dissrepl}{list, string, or object of class \code{u.diss}. If a list, list of same length as \code{MCrseqdata}. List of dissimilarity matrices or \code{dist} objects. If a character string, a method name for computing the dissimilarities with \code{\link{MCudist}}. Can also be an object of class \code{u.diss} previously computed with \code{MCudist}.}

\item{MCrseqdata}{list of MC-replicated sequence datasets of class \code{stslist}. The last element is supposed to be the observed dataset.}

\item{udiss}{logical. When \code{dissrepl} is a distance method, should distance be computed with \code{\link{MCudist}}. See details.}

\item{full.matrix}{logical. Should dissimilarities be organized in matrix form? Default is \code{FALSE} in which case dissimilarity matrices are converted into \code{dist} objects. If \code{TRUE}, dissimilarity \code{dist} objects are converted into matrices.}

\item{...}{additional arguments passed to \code{\link{MCudist}} or \code{\link{MCdisslist}} when \code{dissrepl} is a method name.}
}
\value{
Five objects:\cr
 \code{MCmean} Mean of distance objects over replicated sets of sequences.\cr
 \code{MCsd} Standard deviation of distances over replicated sets of sequences.\cr
 In addition, when the observed distances are provided as last element of the \code{dissrepl} list:\cr
 \code{MCbias} Difference between observed distance and \code{MCmean}\cr
 \code{MCse} Standard error of individual distances.\cr
 \code{MCmse} Mean square error of individual distances.\cr
 The five objects are of class \code{dist} when \code{attr(MCrseqdata,"toref")==FALSE} and matrices otherwise.
}
\description{
Computes the mean and standard deviation of
each element of the pairwise distance matrix
across sets of MC-replicated sequences.
}
\details{
Providing \code{u.diss} distances or computing distances with \code{MCudist} may be faster and can save space when the number of unique replicated sequences is smaller than the sample size times the squared root of R, which can be checked with \code{\link{MCnunique}}. When the number of unique replicated sequences largely exceeds the threshold, it is more efficient to compute distance matrices separately for each updated set of sequences with \code{\link{MCdisslist}} or by setting \code{udiss=FALSE}.
}
\examples{
# example code
exdata <- read.table(text="
                a a b b
                a a b b
                b b a a
                a c c b
                b b a c
                b b a c
                ")
weights=rep(1, nrow(exdata))
s.exdata <- seqdef(exdata, weights = weights, id=paste("id",1:nrow(exdata), sep=""))

## 3 MC-replicated sequence datasets
altseq.list <- MCseqReplicate(s.exdata, J=1, R=3, include.obs=TRUE)
## list of dissimilarity matrices
disslist <- MCdisslist(altseq.list, method="HAM")

MCdselist <- MCseqdistSE(disslist)
print(MCdselist)

MCratioslist <- MCratios(MCdselist)
print(MCratioslist)

}
\seealso{
\code{\link{MCseqReplicate}}, \code{\link{MCdisslist}}, \code{\link{MCudist}}, \code{\link{print.distMC}}, \code{\link{summary.distMC}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCudist.R
\name{MCnunique}
\alias{MCnunique}
\title{Number of unique replicated sequences}
\usage{
MCnunique(MCrseqdata, check = FALSE)
}
\arguments{
\item{MCrseqdata}{list of replicated \code{stslist} state sequence datasets (all of same size and with same alphabet.}

\item{check}{logical. When \code{TRUE}, check if the number of unique replicated sequences is less than  \code{sqrt(number of replicated datasets) * (sample size)}?}
}
\value{
\code{nu} number of unique replicated sequences and, when \code{check=TRUE}, \code{u.ok} the check result.
}
\description{
Number of unique replicated sequences
}
\seealso{
\code{\link{MCudist}}, \code{\link{MCseqdistSE}}
}
